import operator
from typing import AbstractSet

import pytest

from ..config import config
from ..expressions import Constant
from ..probabilistic.cplogic import testing
from ..relational_algebra import (
    ColumnInt,
    ColumnStr,
    Difference,
    ExtendedProjection,
    FunctionApplicationListMember,
    NamedRelationalAlgebraFrozenSet,
    NaturalJoin,
    NumberColumns,
    Projection,
    RelationalAlgebraSet,
    RenameColumn,
    Selection,
    Union,
    int2columnint_constant,
    str2columnstr_constant
)
from ..relational_algebra_provenance import (
    BuildProvenanceAlgebraSetWalkIntoMixin,
    ProvenanceAlgebraSet,
    RelationalAlgebraProvenanceExpressionSemringSolverMixin,
    RelationalAlgebraSolver
)

EQ = Constant(operator.eq)


class RelationalAlgebraProvenanceExpressionSemringSolver(
    BuildProvenanceAlgebraSetWalkIntoMixin,
    RelationalAlgebraProvenanceExpressionSemringSolverMixin,
    RelationalAlgebraSolver
):
    pass


def test_integer_addition_semiring():
    r1 = testing.build_ra_provenance_set_from_named_ra_set(
        NamedRelationalAlgebraFrozenSet(("_p_", "x"), [(2, "a"), (3, "b")]),
        ColumnStr("_p_"),
    )
    r2 = testing.build_ra_provenance_set_from_named_ra_set(
        NamedRelationalAlgebraFrozenSet(("_p_", "x"), [(5, "a"), (10, "c")]),
        ColumnStr("_p_"),
    )
    op = NaturalJoin(r1, r2)
    solver = RelationalAlgebraProvenanceExpressionSemringSolver()
    result = solver.walk(op)

    expected = ProvenanceAlgebraSet(
        Constant[AbstractSet](
            NamedRelationalAlgebraFrozenSet(("_p_", "x"), [(10, "a")])
        ),
        str2columnstr_constant("_p_"),
    )
    assert testing.eq_prov_relations(result, expected)


class SetType(frozenset):
    def __add__(self, other):
        return self.union(other)

    def __mul__(self, other):
        return self.union(other)


@pytest.mark.skipif(
    config["RAS"].get("backend") == "dask",
    reason="multiplication of sets not yet implemented in dask backend",
)
def test_set_type_semiring():
    r1 = testing.build_ra_provenance_set_from_named_ra_set(
        NamedRelationalAlgebraFrozenSet(
            ("_p_", "x"),
            [(SetType({"a", "b"}), "hello"), (SetType({"c", "a"}), "bonjour")],
        ),
        ColumnStr("_p_"),
    )
    r2 = testing.build_ra_provenance_set_from_named_ra_set(
        NamedRelationalAlgebraFrozenSet(
            ("_p_", "x"),
            [(SetType({"c"}), "hello"), (SetType({"c", "a"}), "zoo")],
        ),
        ColumnStr("_p_"),
    )
    op = NaturalJoin(r1, r2)
    solver = RelationalAlgebraProvenanceExpressionSemringSolver()
    result = solver.walk(op)
    expected = ProvenanceAlgebraSet(
        Constant[AbstractSet](NamedRelationalAlgebraFrozenSet(
            ("_p_", "x"), [(SetType({"a", "b", "c"}), "hello")]
        )),
        str2columnstr_constant("_p_"),
    )
    assert testing.eq_prov_relations(result, expected)


class StringTestType:
    def __init__(self, value):
        self.value = value

    def __add__(self, other):
        return StringTestType(self.value + other.value)

    def __mul__(self, other):
        return StringTestType(self.value * len(other.value))

    def __eq__(self, other):
        return self.value == other.value

    def __hash__(self):
        return hash(self.value)


@pytest.mark.skipif(
    config["RAS"].get("backend") == "dask",
    reason="multiplication of strings not yet implemented in dask backend",
)
def test_string_semiring():
    r1 = testing.build_ra_provenance_set_from_named_ra_set(
        NamedRelationalAlgebraFrozenSet(
            ("_p_", "word"),
            [
                (StringTestType("walter"), "say"),
                (StringTestType("white"), "my"),
                (StringTestType("heisenberg"), "name"),
            ],
        ),
        ColumnStr("_p_"),
    )
    r2 = testing.build_ra_provenance_set_from_named_ra_set(
        NamedRelationalAlgebraFrozenSet(
            ("_p_", "word"),
            [(StringTestType("he"), "my"), (StringTestType("the"), "name")],
        ),
        ColumnStr("_p_"),
    )
    op = NaturalJoin(r1, r2)
    solver = RelationalAlgebraProvenanceExpressionSemringSolver()
    result = solver.walk(op)
    expected = ProvenanceAlgebraSet(
        Constant[AbstractSet](NamedRelationalAlgebraFrozenSet(
            ("_p_", "word"),
            [
                (StringTestType("white" * 2), "my"),
                (StringTestType("heisenberg" * 3), "name"),
            ],
        )),
        str2columnstr_constant("_p_"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_multiple_columns():
    r = testing.build_ra_provenance_set_from_named_ra_set(
        NamedRelationalAlgebraFrozenSet(
            ("_p_", "x", "y"),
            [(42, "a", "b"), (21, "a", "z"), (12, "b", "y"), (89, "b", "h"),],
        ),
        ColumnStr("_p_"),
    )
    op = Projection(r, (str2columnstr_constant("x"),))
    solver = RelationalAlgebraProvenanceExpressionSemringSolver()
    result = solver.walk(op)
    expected = ProvenanceAlgebraSet(
        Constant[AbstractSet](NamedRelationalAlgebraFrozenSet(
            ("_p_", "x"), [(63, "a"), (101, "b"),],
        )),
        str2columnstr_constant("_p_"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_renaming():
    r = testing.build_ra_provenance_set_from_named_ra_set(
        NamedRelationalAlgebraFrozenSet(
            ("_p_", "x", "y"),
            [(42, "a", "b"), (21, "a", "z"), (12, "b", "y"), (89, "b", "h"),],
        ),
        ColumnStr("_p_"),
    )
    op = RenameColumn(
        r, str2columnstr_constant("x"), str2columnstr_constant("z"),
    )
    solver = RelationalAlgebraProvenanceExpressionSemringSolver()
    result = solver.walk(op)
    expected = ProvenanceAlgebraSet(
        Constant[AbstractSet](NamedRelationalAlgebraFrozenSet(
            ("_p_", "z", "y"),
            [(42, "a", "b"), (21, "a", "z"), (12, "b", "y"), (89, "b", "h"),],
        )),
        str2columnstr_constant("_p_"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_selection():
    r = testing.build_ra_provenance_set_from_named_ra_set(
        NamedRelationalAlgebraFrozenSet(
            ("_p_", "x", "y"),
            [(42, "a", "b"), (21, "a", "z"), (12, "b", "y"), (89, "b", "h"),],
        ),
        ColumnStr("_p_"),
    )
    op = Selection(
        r, Constant(operator.eq)(str2columnstr_constant("x"), Constant("a"))
    )
    solver = RelationalAlgebraProvenanceExpressionSemringSolver()
    result = solver.walk(op)
    expected = ProvenanceAlgebraSet(
        Constant[AbstractSet](NamedRelationalAlgebraFrozenSet(
            ("_p_", "x", "y"), [(42, "a", "b"), (21, "a", "z")],
        )),
        str2columnstr_constant("_p_"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_union():
    r1 = testing.build_ra_provenance_set_from_named_ra_set(
        NamedRelationalAlgebraFrozenSet(("_p1_", "x"), [(2, "a"), (3, "b")]),
        ColumnStr("_p1_"),
    )
    r2 = testing.build_ra_provenance_set_from_named_ra_set(
        NamedRelationalAlgebraFrozenSet(("_p2_", "x"), [(5, "b"), (10, "c")]),
        ColumnStr("_p2_"),
    )
    op = Union(r1, r2)
    solver = RelationalAlgebraProvenanceExpressionSemringSolver()
    result = solver.walk(op)
    expected = ProvenanceAlgebraSet(
        Constant[AbstractSet](NamedRelationalAlgebraFrozenSet(
            ("_p_", "x"), [(2, "a"), (8, "b"), (10, "c")]
        )),
        str2columnstr_constant("_p_"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_selection_by_columnint():
    r = testing.build_ra_provenance_set_from_named_ra_set(
        NamedRelationalAlgebraFrozenSet(
            ("_p_", "x", "y"), [(0.2, "a", "b"), (0.5, "b", "a")]
        ),
        ColumnStr("_p_"),
    )
    op = Selection(r, EQ(Constant(ColumnInt(0)), Constant("a")))
    solver = RelationalAlgebraProvenanceExpressionSemringSolver()
    result = solver.walk(op)
    expected = ProvenanceAlgebraSet(
        Constant[AbstractSet](
            NamedRelationalAlgebraFrozenSet(
                ("_p_", "x", "y"), [(0.2, "a", "b")]
            )
        ),
        str2columnstr_constant("_p_"),
    )
    assert testing.eq_prov_relations(result, expected)
    op = Selection(r, EQ(Constant(ColumnInt(0)), Constant("a")))
    op = Selection(op, EQ(Constant(ColumnInt(1)), Constant("b")))
    solver = RelationalAlgebraProvenanceExpressionSemringSolver()
    result = solver.walk(op)
    expected = ProvenanceAlgebraSet(
        Constant[AbstractSet](NamedRelationalAlgebraFrozenSet(
            ("_p_", "x", "y"),
            [(0.2, "a", "b")])
        ),
        str2columnstr_constant("_p_"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_projection_columnint():
    r = testing.build_ra_provenance_set_from_named_ra_set(
        NamedRelationalAlgebraFrozenSet(
            ("_p_", "x", "y"), [(0.2, "a", "b"), (0.5, "b", "a")]
        ),
        ColumnStr("_p_"),
    )

    r = NumberColumns(r, (str2columnstr_constant("x"), str2columnstr_constant("y")))
    op = Projection(r, (Constant(ColumnInt(0)),))
    solver = RelationalAlgebraProvenanceExpressionSemringSolver()
    result = solver.walk(op)
    expected = ProvenanceAlgebraSet(
        Constant[AbstractSet](RelationalAlgebraSet(
            [(0.2, "a"), (0.5, "b")]
        )),
        int2columnint_constant(0),
    )
    assert result == expected
    op = Projection(r, (Constant(ColumnInt(0)), Constant(ColumnInt(1))))
    solver = RelationalAlgebraProvenanceExpressionSemringSolver()
    result = solver.walk(op)
    assert result == solver.walk(r)
    op = Projection(r, (Constant(ColumnInt(1)), Constant(ColumnInt(0))))
    solver = RelationalAlgebraProvenanceExpressionSemringSolver()
    result = solver.walk(op)
    expected = ProvenanceAlgebraSet(
        Constant[AbstractSet](RelationalAlgebraSet(
            [(0.2, "b", "a"), (0.5, "a", "b")]
        )),
        int2columnint_constant(0)
    )
    assert result == expected


def test_selection_between_columnints():
    r = testing.build_ra_provenance_set_from_named_ra_set(
        NamedRelationalAlgebraFrozenSet(
            columns=("_p_", "x", "y"),
            iterable=[
                (0.5, "a", "b"),
                (0.7, "a", "a"),
                (0.2, "b", "b"),
            ],
        ),
        ColumnStr("_p_"),
    )
    col1 = Constant[ColumnInt](ColumnInt(0))
    col2 = Constant[ColumnInt](ColumnInt(1))
    op = Selection(r, Constant(operator.eq)(col1, col2))
    solver = RelationalAlgebraProvenanceExpressionSemringSolver()
    result = solver.walk(op)
    expected = ProvenanceAlgebraSet(
        Constant[AbstractSet](NamedRelationalAlgebraFrozenSet(
            columns=("_p_", "x", "y"),
            iterable=[
                (0.7, "a", "a"),
                (0.2, "b", "b"),
            ],
        )),
        str2columnstr_constant("_p_"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_difference():
    r_left = testing.build_ra_provenance_set_from_named_ra_set(
        NamedRelationalAlgebraFrozenSet(
            columns=("_p1_", "x", "y", "w"),
            iterable=[
                (0.5, "a", "b", "a"),
                (0.7, "a", "a", "b"),
                (0.2, "b", "b", "c"),
            ],
        ),
        ColumnStr("_p1_"),
    )

    r_right = testing.build_ra_provenance_set_from_named_ra_set(
        NamedRelationalAlgebraFrozenSet(
            columns=("_p2_", "x", "y", "z"),
            iterable=[
                (0.2, "a", "b", "a"),
                (0.3, "b", "a", "b"),
                (0.1, "b", "b", "c"),
            ],
        ),
        ColumnStr("_p2_"),
    )

    r_expected = ProvenanceAlgebraSet(
        Constant[AbstractSet](NamedRelationalAlgebraFrozenSet(
            columns=("_p1_", "x", "y", "w"),
            iterable=[
                (0.4, "a", "b", "a"),
                (0.7, "a", "a", "b"),
                (0.18, "b", "b", "c"),
            ],
        )),
        str2columnstr_constant("_p1_"),
    )

    op = Projection(
        Difference(r_left, r_right),
        tuple(str2columnstr_constant(c) for c in ("x", "y", "w"))
    )
    result = RelationalAlgebraProvenanceExpressionSemringSolver().walk(op)
    assert testing.eq_prov_relations(result, r_expected)


def test_difference_same_provenance_column():
    r_left = testing.build_ra_provenance_set_from_named_ra_set(
        NamedRelationalAlgebraFrozenSet(
            columns=("_p_", "x", "y", "w"),
            iterable=[
                (0.5, "a", "b", "a"),
                (0.3, "a", "a", "b"),
                (0.2, "b", "b", "c"),
            ],
        ),
        ColumnStr("_p_"),
    )

    r_right = testing.build_ra_provenance_set_from_named_ra_set(
        NamedRelationalAlgebraFrozenSet(
            columns=("_p_", "x", "y", "z"),
            iterable=[
                (0.7, "a", "b", "a"),
                (0.8, "b", "a", "b"),
                (0.9, "b", "b", "c"),
            ],
        ),
        ColumnStr("_p_"),
    )

    r_expected = ProvenanceAlgebraSet(
        Constant[AbstractSet](NamedRelationalAlgebraFrozenSet(
            columns=("_p_", "x", "y", "w"),
            iterable=[
                (0.15, "a", "b", "a"),
                (0.3, "a", "a", "b"),
                (0.02, "b", "b", "c"),
            ],
        )),
        str2columnstr_constant("_p_"),
    )

    op = Projection(
        Difference(r_left, r_right),
        tuple(str2columnstr_constant(c) for c in ("x", "y", "w"))
    )
    result = RelationalAlgebraProvenanceExpressionSemringSolver().walk(op)
    assert testing.eq_prov_relations(result, r_expected)


def test_extended_proj():
    provset = testing.build_ra_provenance_set_from_named_ra_set(
        NamedRelationalAlgebraFrozenSet(
            ("_p_", "x", "y"),
            [
                (0.2, "a", "b"),
                (0.3, "b", "a"),
                (0.5, "c", "c"),
            ],
        ),
        ColumnStr("_p_"),
    )
    proj_list = [
        FunctionApplicationListMember(
            str2columnstr_constant("x"), str2columnstr_constant("x")
        ),
        FunctionApplicationListMember(
            str2columnstr_constant("y"), str2columnstr_constant("y")
        ),
        FunctionApplicationListMember(
            Constant("d"), str2columnstr_constant("z")
        ),
    ]
    proj = ExtendedProjection(provset, proj_list)
    solver = RelationalAlgebraProvenanceExpressionSemringSolver()
    result = solver.walk(proj)
    expected = ProvenanceAlgebraSet(
        Constant[AbstractSet](NamedRelationalAlgebraFrozenSet(
            ("_p_", "x", "y", "z"),
            [
                (0.2, "a", "b", "d"),
                (0.3, "b", "a", "d"),
                (0.5, "c", "c", "d"),
            ],
        )),
        str2columnstr_constant("_p_"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_forbidden_extended_proj_missing_nonprov_cols():
    provset = testing.build_ra_provenance_set_from_named_ra_set(
        NamedRelationalAlgebraFrozenSet(
            ("_p_", "x", "y"),
            [
                (0.2, "a", "b"),
                (0.3, "b", "a"),
                (0.5, "c", "c"),
            ],
        ),
        ColumnStr("_p_"),
    )
    proj_list = [
        FunctionApplicationListMember(
            str2columnstr_constant("x"), str2columnstr_constant("x")
        ),
        FunctionApplicationListMember(
            Constant("d"), str2columnstr_constant("z")
        ),
    ]
    proj = ExtendedProjection(provset, proj_list)
    solver = RelationalAlgebraProvenanceExpressionSemringSolver()
    with pytest.raises(ValueError):
        solver.walk(proj)


def test_forbidden_extended_proj_on_provcol():
    provset = testing.build_ra_provenance_set_from_named_ra_set(
        NamedRelationalAlgebraFrozenSet(
            ("_p_", "x", "y"),
            [
                (0.2, "a", "b"),
                (0.3, "b", "a"),
                (0.5, "c", "c"),
            ],
        ),
        ColumnStr("_p_"),
    )
    proj_list = [
        FunctionApplicationListMember(
            str2columnstr_constant("x"), str2columnstr_constant("x")
        ),
        FunctionApplicationListMember(
            str2columnstr_constant("y"), str2columnstr_constant("y")
        ),
        FunctionApplicationListMember(
            Constant("d"), str2columnstr_constant("_p_")
        ),
    ]
    proj = ExtendedProjection(provset, proj_list)
    solver = RelationalAlgebraProvenanceExpressionSemringSolver()
    with pytest.raises(ValueError):
        solver.walk(proj)



import numpy as np

from ..probabilistic_frontend import NeurolangPDL


def assert_almost_equal(set_a, set_b):
    for tupl_a in set_a:
        assert any(
            all(
                np.isclose(term_a, term_b)
                if isinstance(term_a, (float, np.float32, np.float64))
                else term_a == term_b
                for term_a, term_b in zip(tupl_a, tupl_b)
            )
            for tupl_b in set_b
        )


def test_deterministic_segregation_query():
    nl = NeurolangPDL()
    nl.add_tuple_set(
        [("a",), ("b",), ("c",)],
        name="Network",
    )
    nl.add_tuple_set(
        [("s1",), ("s2",), ("s3",)],
        name="Study",
    )
    nl.add_tuple_set(
        [
            ("a", "s1"),
            ("b", "s1"),
            ("b", "s2"),
            ("c", "s2"),
            ("c", "s3"),
        ],
        name="NetworkReported",
    )
    with nl.environment as e:
        e.SegregationQuery[e.n, e.s] = e.NetworkReported(
            e.n, e.s
        ) & ~nl.exists(e.n2, e.NetworkReported(e.n2, e.s) & (e.n2 != e.n))
        result = nl.query((e.n, e.s), e.SegregationQuery(e.n, e.s))
    expected = {
        ("c", "s3"),
    }
    assert_almost_equal(result, expected)



import operator
from typing import AbstractSet, Tuple
from unittest.mock import Mock

import pytest

from .....datalog import Fact
from .....datalog.expression_processing import extract_logic_atoms
from .....exceptions import ForbiddenExpressionError
from .....expression_walker import ExpressionWalker, IdentityWalker
from .....expressions import Constant, Symbol
from .....logic import Conjunction, Implication
from .....probabilistic.expressions import (
    PROB,
    ProbabilisticFact,
    ProbabilisticQuery,
)
from ... import sugar


class SymbolTableMixin:
    def __init__(self, symbol_table=None):
        if symbol_table is None:
            symbol_table = dict()
        self.symbol_table = symbol_table


class TranslateColumnsToAtoms(
    sugar.TranslateColumnsToAtoms, SymbolTableMixin, ExpressionWalker
):
    pass


class TranslateSelectByFirstColumn(
    sugar.TranslateSelectByFirstColumn, SymbolTableMixin, ExpressionWalker
):
    pass


def test_columns_to_atoms_rules():
    A = Symbol("A")
    B = Symbol("B")
    C = Symbol("C")
    x = Symbol("x")
    one = Constant(1)

    test_rule = Implication(A(x), C(sugar.Column(B, one), x))

    tcta = TranslateColumnsToAtoms({B: Constant(Mock(arity=3))})
    out_rule = tcta.walk(test_rule)
    antecedent_formulas = out_rule.antecedent.formulas
    if antecedent_formulas[0].functor == "C":
        c = antecedent_formulas[0]
        b = antecedent_formulas[1]
    else:
        c = antecedent_formulas[1]
        b = antecedent_formulas[0]

    assert isinstance(out_rule.antecedent, Conjunction)
    assert c.functor == C
    assert c.args[1] == x
    assert all(arg.is_fresh for arg in b.args)
    assert c.args[0] == b.args[1]

    test_rule = Implication(
        A(sugar.Column(B, one)), C(sugar.Column(B, one), x)
    )

    tcta = TranslateColumnsToAtoms({B: Constant(Mock(arity=3))})
    out_rule = tcta.walk(test_rule)
    antecedent_formulas = out_rule.antecedent.formulas
    if antecedent_formulas[0].functor == "C":
        c = antecedent_formulas[0]
        b = antecedent_formulas[1]
    else:
        c = antecedent_formulas[1]
        b = antecedent_formulas[0]

    assert isinstance(out_rule.antecedent, Conjunction)
    assert c.functor == C
    assert c.args[1] == x
    assert all(arg.is_fresh for arg in b.args)
    assert c.args[0] == b.args[1]
    assert out_rule.consequent.functor == A
    assert out_rule.consequent.args == (b.args[1],)

    test_rule = Implication(A(x), Conjunction((B(x, x, x), C(x))))
    out_rule = tcta.walk(test_rule)

    assert test_rule == out_rule


def test_columns_to_atoms_facts():
    A = Symbol("A")
    B = Symbol("B")
    one = Constant(1)

    test_rule = Fact(A(sugar.Column(B, one)))

    tcta = TranslateColumnsToAtoms({B: Constant(Mock(arity=3))})
    out_rule = tcta.walk(test_rule)
    atoms = extract_logic_atoms(out_rule.antecedent)
    assert len(atoms) == 1
    b = atoms.pop()
    assert b.functor == B
    assert all(arg.is_fresh for arg in b.args)
    assert out_rule.consequent.functor == A
    assert out_rule.consequent.args[0] == b.args[1]


def test_select_by_first_implication():
    A = Symbol("A")
    B = Symbol("B")
    C = Symbol("C")
    c = Constant("c")
    x = Symbol("x")

    test_rule = Implication(A(x), C(sugar.SelectByFirstColumn(B, c), x))

    tr = TranslateSelectByFirstColumn().walk(test_rule)
    fs = next(s for s in tr._symbols if s.is_fresh)
    assert tr == Implication(A(x), Conjunction((C(fs, x), B(c, fs))))


def test_select_by_first_implication_builtin():
    A = Symbol("A")
    B = Symbol("B")
    C = Symbol("C")
    c = Constant("c")
    eq = Constant(lambda x, y: x == y)
    x = Symbol("x")
    y = Symbol("y")

    test_rule = Implication(
        A(x),
        Conjunction(
            (
                C(sugar.SelectByFirstColumn(B, c), x),
                eq(sugar.SelectByFirstColumn(B, c), y),
            )
        ),
    )

    tr = TranslateSelectByFirstColumn().walk(test_rule)
    fs = next(s for s in tr._symbols if s.is_fresh)
    res = Implication(A(x), Conjunction((C(fs, x), eq(fs, y), B(c, fs))))
    assert tr == res


def test_select_by_first_implication_builtin_head():
    A = Symbol("A")
    B = Symbol("B")
    C = Symbol[AbstractSet[Tuple[int, int]]]("C")
    c = Constant("c")
    eq = Constant(lambda x, y: x == y)
    x = Symbol("x")
    y = Symbol("y")

    test_rule = Implication(
        sugar.SelectByFirstColumn(A, c), Conjunction((C(x), eq(y), B(x)))
    )

    tr = TranslateSelectByFirstColumn({C: C, B: B}).walk(test_rule)
    fresh_symbols = [s for s in tr._symbols if s.is_fresh]
    assert len(fresh_symbols) == 1

    fs = fresh_symbols[0]
    res = Implication(A(c, fs), Conjunction((C(fs, x), eq(fs, y), B(x))))
    assert tr == res


class _TestTranslator(
    sugar.TranslateProbabilisticQueryMixin, ExpressionWalker
):
    pass


def test_wlq_floordiv_translation():
    P = Symbol("P")
    Q = Symbol("Q")
    R = Symbol("R")
    x = Symbol("x")
    y = Symbol("y")
    EQ_ = Constant(operator.eq)

    wlq = Implication(
        Q(x, y, PROB(x, y)), Constant(operator.floordiv)(P(x), R(x, y))
    )
    translator = _TestTranslator()
    result = translator.walk(wlq)

    # assert num == fresh_01(x, y, PROB) :- P(x) & R(x, y)
    assert len(result) == 3
    num = result[0]
    fnum = num.consequent.functor
    assert fnum.is_fresh
    assert num == Implication(
        fnum(x, y, ProbabilisticQuery(PROB, (x, y))),
        Conjunction((P(x), R(x, y))),
    )

    # assert denum == fresh_02(x, y, PROB) :- R(x, y)
    denum = result[1]
    fdenum = denum.consequent.functor
    assert fdenum.is_fresh
    assert denum == Implication(
        fdenum(x, y, ProbabilisticQuery(PROB, (x, y))), R(x, y)
    )

    # assert cond == Q(x, y, p) :- fresh_01(x, y, p0) & fresh_02(x, y, p1) & (p == p0 / p1)
    cond = result[2]
    p = [a for a in cond.consequent.args if a.is_fresh][0]
    p0 = [a for a in cond.antecedent.formulas[0].args if a.is_fresh][0]
    p1 = [a for a in cond.antecedent.formulas[1].args if a.is_fresh][0]

    assert cond == Implication(
        Q(x, y, p),
        Conjunction(
            (
                fnum(x, y, p0),
                fdenum(x, y, p1),
                EQ_(p, Constant(operator.truediv)(p0, p1)),
            )
        ),
    )


def test_wlq_floordiv_translation_boolean_denominator():
    P = Symbol("P")
    Q = Symbol("Q")
    R = Symbol("R")
    x = Symbol("x")
    y = Symbol("y")
    EQ_ = Constant(operator.eq)

    wlq = Implication(
        Q(x, PROB(x)), Constant(operator.floordiv)(P(x, y), R(y))
    )
    translator = _TestTranslator()
    result = translator.walk(wlq)

    # assert num == fresh_01(x, PROB) :- P(x, y) & R(y)
    assert len(result) == 3
    num = result[0]
    fnum = num.consequent.functor
    assert fnum.is_fresh
    assert num == Implication(
        fnum(x, ProbabilisticQuery(PROB, (x,))),
        Conjunction((P(x, y), R(y))),
    )

    # assert denum == fresh_02(PROB) :- R(y)
    denum = result[1]
    fdenum = denum.consequent.functor
    assert fdenum.is_fresh
    assert denum == Implication(
        fdenum(ProbabilisticQuery(PROB, tuple())), R(y)
    )

    # assert cond == Q(x, y, p) :- fresh_01(x, y, p0) & fresh_02(x, y, p1) & (p == p0 / p1)
    cond = result[2]
    p = [a for a in cond.consequent.args if a.is_fresh][0]
    p0 = [a for a in cond.antecedent.formulas[0].args if a.is_fresh][0]
    p1 = [a for a in cond.antecedent.formulas[1].args if a.is_fresh][0]

    assert cond == Implication(
        Q(x, p),
        Conjunction(
            (
                fnum(x, p0),
                fdenum(p1),
                EQ_(p, Constant(operator.truediv)(p0, p1)),
            )
        ),
    )


def test_wlq_marg_bad_syntax():
    P = Symbol("P")
    Q = Symbol("Q")
    Z = Symbol("Z")
    x = Symbol("x")
    y = Symbol("y")
    bad_wlq = Implication(Q(x, y), Constant(operator.floordiv)(P(x), Z(x, y)))
    translator = _TestTranslator()
    with pytest.raises(ForbiddenExpressionError):
        translator.walk(bad_wlq)


class TestTranslateQueryBasedProbabilisticFact(
    sugar.TranslateQueryBasedProbabilisticFactMixin,
    IdentityWalker,
):
    pass


def test_translation_sugar_syntax():
    P = Symbol("P")
    Q = Symbol("Q")
    x = Symbol("x")
    p = Symbol("p")
    pfact = Implication(
        Constant(operator.matmul)(P, (p / Constant(2)))(x), Q(x, p)
    )
    translator = TestTranslateQueryBasedProbabilisticFact()
    result = translator.walk(pfact)
    expected = Implication(
        ProbabilisticFact(p / Constant(2), P(x)), Q(x, p)
    )
    assert result == expected



from collections import defaultdict
from itertools import product

from ..expression_walker import ReplaceExpressionWalker
from ..expressions import Symbol
from ..logic.expression_processing import extract_logic_atoms
from ..logic.transformations import convert_to_pnf_with_dnf_matrix
from ..relational_algebra import (
    Projection,
    Selection,
    eq_,
    int2columnint_constant
)


def verify_that_the_query_is_ranked(query):
    argument_precedence = defaultdict(set)
    query = convert_to_pnf_with_dnf_matrix(query)

    for atom in extract_logic_atoms(query):
        if not isinstance(atom.functor, Symbol):
            return False
        for i, arg in enumerate(atom.args[1:], 1):
            argument_precedence[arg].update(atom.args[:i])

    for arg1, arg2 in product(argument_precedence, argument_precedence):
        if (
            arg1 in argument_precedence[arg2] and
            arg2 in argument_precedence[arg1]
        ):
            return False
    return True


def partially_rank_query(query, symbol_table):
    atoms_to_rank, atom_argument_quantity = \
        _process_atoms_with_repeatd_arguments(query)

    for symbol, instances_to_rank in atoms_to_rank.items():
        if len(instances_to_rank) != 1:
            continue
        arguments_to_rank = instances_to_rank.pop()
        if all(len(repeats) == 0 for _, repeats in arguments_to_rank):
            continue
        expression, arguments_to_eliminate = _build_ra_expression(
            atom_argument_quantity, symbol, arguments_to_rank
        )
        new_symbol = Symbol.fresh()
        symbol_table[new_symbol] = expression
        replacements = {
            atom: new_symbol(*(
                arg for i, arg in enumerate(atom.args)
                if i not in arguments_to_eliminate
            )
            )
            for atom in extract_logic_atoms(query)
            if atom.functor == symbol
        }
        query = ReplaceExpressionWalker(replacements).walk(query)
    return query


def _build_ra_expression(atom_argument_quantity, symbol, arguments_to_rank):
    expression = symbol
    arguments_to_eliminate = set()
    for arg, repeats in arguments_to_rank:
        arg = int2columnint_constant(arg)
        for repeat in repeats:
            expression = Selection(
                    expression,
                    eq_(arg, int2columnint_constant(repeat))
                )
        arguments_to_eliminate.update(repeats)
    attributes = tuple(
            int2columnint_constant(i)
            for i in range(atom_argument_quantity[symbol])
            if i not in arguments_to_eliminate
        )
    expression = Projection(expression, attributes)
    return expression, arguments_to_eliminate


def _process_atoms_with_repeatd_arguments(query):
    atoms_to_rank = defaultdict(set)
    atom_argument_quantity = dict()
    for atom in extract_logic_atoms(query):
        if not isinstance(atom.functor, Symbol):
            continue
        atom_argument_quantity[atom.functor] = len(atom.args)
        arg_counts = defaultdict(list)
        for i, arg in enumerate(atom.args):
            arg_counts[arg].append(i)
        arg_counts_tuple = tuple(
            (v[0], tuple(v[1:])) for v in arg_counts.values()
        )
        atoms_to_rank[atom.functor].add(arg_counts_tuple)
    return atoms_to_rank, atom_argument_quantity



import operator

import pandas as pd
import pytest
from pytest import raises


from ... import expression_walker, expressions
from ...frontend.datalog.sugar import TranslateProbabilisticQueryMixin
from ...frontend.probabilistic_frontend import NeurolangPDL
from ...logic import Negation
from .. import DatalogProgram, Fact, Implication, Conjunction
from ..aggregation import (
    AGG_COUNT,
    BuiltinAggregationMixin,
    DatalogWithAggregationMixin,
    TranslateToLogicWithAggregation,
)
from ..chase import Chase
from ..exceptions import (
    BoundAggregationApplicationError,
    NegationInMagicSetsRewriteError,
)
from ..magic_sets import (
    AdornedSymbol,
    LeftToRightSIPS,
    ReplaceAdornedSymbolWalker,
    magic_rewrite,
)
from ..negation import DatalogProgramNegationMixin

C_ = expressions.Constant
S_ = expressions.Symbol
Imp_ = Implication
Cnj_ = Conjunction
F_ = Fact
Eb_ = expressions.ExpressionBlock


class Datalog(
    TranslateProbabilisticQueryMixin,
    TranslateToLogicWithAggregation,
    BuiltinAggregationMixin,
    DatalogWithAggregationMixin,
    DatalogProgramNegationMixin,
    DatalogProgram,
    expression_walker.ExpressionBasicEvaluator,
):
    pass


x = S_("x")
y = S_("y")
z = S_("z")
x1 = S_("x1")
y1 = S_("y1")
rsg = S_("rsg")
up = S_("up")
down = S_("down")
flat = S_("flat")
anc = S_("anc")
par = S_("par")
anc2 = S_("anc2")
q = S_("q")
a = C_("a")
b = C_("b")
c = C_("c")
d = C_("d")
eq = C_(operator.eq)

aq = AdornedSymbol(q, "f", 0)
aanc2 = AdornedSymbol(anc2, "bf", 0)
aanc = AdornedSymbol(anc, "bf", 0)


def test_adorned_symbol():
    ax = AdornedSymbol(x, 'b', 1)
    ax_ = AdornedSymbol(x, 'b', 1)
    ax__ = AdornedSymbol(a, 'b', 1)

    assert ax.expression == x
    assert ax.adornment == 'b'
    assert ax.number == 1
    assert ax == ax_
    assert ax != ax__

    ax = AdornedSymbol(a, 'b', 1)
    ax_ = AdornedSymbol(a, 'b', 1)
    ax__ = AdornedSymbol(a, 'b', 2)

    assert ax.expression == a
    assert ax.adornment == 'b'
    assert ax.number == 1
    assert ax == ax_
    assert ax != ax__

    with raises(NotImplementedError):
        AdornedSymbol(a, 'b', 1).name


def test_unadorned_expression_walker():
    adorned_code = tuple(
        [
            Imp_(aq(x), aanc2(a, x)),
            Imp_(aanc(x, y), par(x, y)),
            Imp_(aanc(x, y), aanc(x, z) & par(z, y)),
            Imp_(aanc2(x, y), aanc(x, y) & Negation(aanc(x, y))),
        ]
    )
    unadorned_code = ReplaceAdornedSymbolWalker().walk(adorned_code)

    xanc2 = S_("anc2^bf_0")
    xanc = S_("anc^bf_0")
    xq = S_("q^f_0")
    expected_code = tuple(
        [
            Imp_(xq(x), xanc2(a, x)),
            Imp_(xanc(x, y), par(x, y)),
            Imp_(xanc(x, y), xanc(x, z) & par(z, y)),
            Imp_(xanc2(x, y), xanc(x, y) & Negation(xanc(x, y))),
        ]
    )
    assert unadorned_code == expected_code


def test_l2r_sips():
    edb = Eb_(
        [
            F_(par(a, b)),
        ]
    )
    dl = Datalog()
    dl.walk(edb)
    sips = LeftToRightSIPS(dl)

    # Default case
    arcs, _ = sips.creates_arcs(Imp_(q(x), anc(a, x)), aq(x))
    assert arcs == {aanc(a, x): (aq(x),)}

    # par is in EDB and should not be an arc
    arcs, _ = sips.creates_arcs(Imp_(anc(x, y), par(x, y)), aanc(x, y))
    assert arcs == {}

    # par is in EDB and should not be an arc
    arcs, _ = sips.creates_arcs(
        Imp_(anc(x, y), Cnj_((anc(x, z), par(z, y)))), aanc(x, y)
    )
    assert arcs == {aanc(x, z): (aanc(x, y),)}

    # bindings should propagate from predicate to predicate
    arcs, _ = sips.creates_arcs(
        Imp_(anc(x, y), Cnj_((anc(y, x), par(z, y), up(y, z), down(z, x)))),
        aanc(x, y),
    )
    ancfb = AdornedSymbol(anc, "fb", 0)(y, x)
    upbf = AdornedSymbol(up, "bf", 0)(y, z)
    downbb = AdornedSymbol(down, "fb", 0)(z, x)
    assert arcs == {
        ancfb: (aanc(x, y),),
        upbf: (aanc(x, y), ancfb),
        downbb: (aanc(x, y), ancfb, upbf),
    }

    # negative predicates should not be added to the tails of arcs
    arcs, _ = sips.creates_arcs(
        Imp_(anc2(x, y), Cnj_((Negation(anc(x, y)), anc2(y, x)))), aanc2(x, y)
    )
    anc2fb = AdornedSymbol(anc2, "fb", 0)
    assert arcs == {
        Negation(aanc(x, y)): (aanc2(x, y),),
        anc2fb(y, x): (aanc2(x, y),),
    }


def test_resolution_works():
    edb = Eb_(
        [
            F_(par(a, b)),
            F_(par(b, c)),
            F_(par(c, d)),
        ]
    )

    code = Eb_(
        [
            Imp_(q(x), anc(a, x)),
            Imp_(anc(x, y), par(x, y)),
            Imp_(anc(x, y), anc(x, z) & par(z, y)),
        ]
    )

    dl = Datalog()
    dl.walk(code)
    dl.walk(edb)
    goal, mr = magic_rewrite(q(x), dl)

    dl = Datalog()
    dl.walk(mr)
    dl.walk(edb)

    solution = Chase(dl).build_chase_solution()
    assert solution[goal].value == {C_((e,)) for e in (b, c, d)}


def test_resolution_works_2():
    """
    Reverse Same Generation from Abiteboul et al. 1995.
    Foundations of databases. p.312

    l0:  l  m   n   o   p
    l1:  e  f   g   h   i   j   k
    l2:  a  b   c   d
    """
    edb = Eb_(
        [
            F_(up(C_("a"), C_("e"))),
            F_(up(C_("a"), C_("f"))),
            F_(up(C_("f"), C_("m"))),
            F_(up(C_("g"), C_("n"))),
            F_(up(C_("h"), C_("n"))),
            F_(up(C_("i"), C_("o"))),
            F_(up(C_("j"), C_("o"))),
            F_(flat(C_("g"), C_("f"))),
            F_(flat(C_("m"), C_("n"))),
            F_(flat(C_("m"), C_("o"))),
            F_(flat(C_("p"), C_("m"))),
            F_(down(C_("l"), C_("f"))),
            F_(down(C_("m"), C_("f"))),
            F_(down(C_("g"), C_("b"))),
            F_(down(C_("h"), C_("c"))),
            F_(down(C_("i"), C_("d"))),
            F_(down(C_("p"), C_("k"))),
        ]
    )

    code = Eb_(
        [
            Imp_(q(y), rsg(C_("a"), y)),
            Imp_(rsg(x, y), flat(x, y)),
            Imp_(rsg(x, y), up(x, x1) & rsg(y1, x1) & down(y1, y)),
        ]
    )

    dl = Datalog()
    dl.walk(code)
    dl.walk(edb)
    goal, mr = magic_rewrite(q(x), dl)

    dl = Datalog()
    dl.walk(mr)
    dl.walk(edb)

    solution = Chase(dl).build_chase_solution()
    assert solution[goal].value == {(C_("b"),), (C_("c"),), (C_("d"),)}


def test_resolution_works_query_constant():
    edb = Eb_(
        [
            F_(par(a, b)),
            F_(par(b, c)),
            F_(par(c, d)),
        ]
    )

    code = Eb_(
        [
            Imp_(q(x, a), anc(a, x)),
            Imp_(anc(x, y), par(x, y)),
            Imp_(anc(x, y), anc(x, z) & par(z, y)),
        ]
    )

    dl = Datalog()
    dl.walk(code)
    dl.walk(edb)
    goal, mr = magic_rewrite(q(x), dl)

    dl = Datalog()
    dl.walk(mr)
    dl.walk(edb)

    solution = Chase(dl).build_chase_solution()
    assert solution[goal].value == {C_((e, a)) for e in (b, c, d)}


def test_resolution_works_query_free():
    edb = Eb_(
        [
            F_(par(a, b)),
            F_(par(b, c)),
            F_(par(c, d)),
        ]
    )

    code = Eb_(
        [
            Imp_(q(x), anc2(x)),
            Imp_(anc(x, y), par(x, y)),
            Imp_(anc(x, y), anc(x, z) & par(z, y)),
            Imp_(anc2(x), anc(a, x)),
        ]
    )

    dl = Datalog()
    dl.walk(code)
    dl.walk(edb)
    goal, mr = magic_rewrite(q(x), dl)

    dl = Datalog()
    dl.walk(mr)
    dl.walk(edb)

    solution = Chase(dl).build_chase_solution()
    assert solution[goal].value == {C_((e,)) for e in (b, c, d)}


def test_resolution_works_builtin():
    edb = Eb_(
        [
            F_(par(a, b)),
            F_(par(b, c)),
            F_(par(c, d)),
        ]
    )

    code = Eb_(
        [
            Imp_(q(x), anc2(a, x)),
            Imp_(anc(x, y), par(x, y)),
            Imp_(anc(x, y), anc(x, z) & par(z, y)),
            Imp_(anc2(x, y), anc(x, z) & eq(z, y)),
        ]
    )

    dl = Datalog()
    dl.walk(code)
    dl.walk(edb)
    goal, mr = magic_rewrite(q(x), dl)

    dl = Datalog()
    dl.walk(mr)
    dl.walk(edb)

    solution = Chase(dl).build_chase_solution()
    assert solution[goal].value == {C_((e,)) for e in (b, c, d)}


def test_resolution_works_aggregation():
    edb = Eb_(
        [
            F_(par(a, b)),
            F_(par(b, c)),
            F_(par(c, d)),
        ]
    )

    code = Eb_(
        [
            Imp_(q(x), anc2(a, x)),
            Imp_(anc(x, y), par(x, y)),
            Imp_(anc(x, y), anc(x, z) & par(z, y)),
            Imp_(anc2(x, AGG_COUNT(y)), anc(x, y)),
        ]
    )

    dl = Datalog()
    dl.walk(code)
    dl.walk(edb)
    goal, mr = magic_rewrite(q(x), dl)

    dl = Datalog()
    dl.walk(mr)
    dl.walk(edb)

    solution = Chase(dl).build_chase_solution()
    assert solution[goal].value == {(3,)}


def test_bound_aggregation_raises_error():
    edb = Eb_(
        [
            F_(par(a, b)),
            F_(par(b, c)),
            F_(par(c, d)),
        ]
    )

    code = Eb_(
        [
            Imp_(q(x), anc2(x, C_(3))),
            Imp_(anc(x, y), par(x, y)),
            Imp_(anc(x, y), anc(x, z) & par(z, y)),
            Imp_(anc2(x, AGG_COUNT(y)), anc(x, y)),
        ]
    )

    dl = Datalog()
    dl.walk(code)
    dl.walk(edb)
    with raises(BoundAggregationApplicationError):
        magic_rewrite(q(x), dl)


def test_resolution_works_negation():
    edb = Eb_(
        [
            F_(par(a, b)),
            F_(par(b, c)),
            F_(par(c, d)),
        ]
    )

    code = Eb_(
        [
            Imp_(q(x), anc2(a, x)),
            Imp_(anc(x, y), par(x, y)),
            Imp_(anc(x, y), anc(x, z) & par(z, y)),
            Imp_(up(x, y), par(x, y)),
            Imp_(anc2(x, y), anc(x, y) & ~up(x, y)),
        ]
    )

    dl = Datalog()
    dl.walk(code)
    dl.walk(edb)

    goal, mr = magic_rewrite(q(x), dl)

    dl = Datalog()
    dl.walk(mr)
    dl.walk(edb)

    solution = Chase(dl).build_chase_solution()
    assert solution[goal].value == {C_((e,)) for e in (c, d)}


def test_complex_negations_raise_errors():
    edb = Eb_(
        [
            F_(par(a, b)),
            F_(par(b, c)),
            F_(par(c, d)),
        ]
    )

    code = Eb_(
        [
            Imp_(q(x), anc2(a, x)),
            Imp_(anc(x, y), par(x, y)),
            Imp_(anc(x, y), anc(x, z) & par(z, y)),
            Imp_(anc2(x, y), anc(x, y) & ~(par(y, z) & anc(x, z))),
        ]
    )

    dl = Datalog()
    dl.walk(code)
    dl.walk(edb)

    with raises(NegationInMagicSetsRewriteError):
        magic_rewrite(q(x), dl)


@pytest.fixture
def nl():
    nl = NeurolangPDL()
    studies = (
        10191322,
        10227106,
        10349031,
        10402199,
        10571235,
        10407201,
        10197540,
        10022496,
        10022494,
        9819274,
        11430815,
        11275483,
        10913505,
        10808134,
        10751455,
    )
    nl.add_uniform_probabilistic_choice_over_set(
        list(map(lambda x: (x,), studies)), name="SelectedStudy"
    )
    peaks_data = (
        (-48, -74, 34, 10349031),
        (-33, 28, 7, 10191322),
        (21, 52, 20, 10349031),
        (-42, 28, 9, 10191322),
        (42, -24, 12, 10022496),
        (-38, 15, 26, 10191322),
        (-36, 26, 9, 10191322),
        (-53, -26, -21, 10349031),
        (-42, 12, 17, 10191322),
        (0, -73, 6, 10913505),
        (10, 41, 1, 10349031),
        (-40, 22, 32, 10191322),
        (-52, -28, 52, 9819274),
        (-48, -30, 12, 10022496),
        (-44, 27, 18, 10191322),
        (-48, -24, 12, 10022496),
        (-30, -31, -2, 10197540),
        (-35, 16, 39, 10191322),
        (18, -76, -16, 10022494),
        (-48, 14, 36, 10227106),
        (-48, 32, -16, 10808134),
        (-48, 26, 27, 10402199),
        (38, -84, 0, 10751455),
        (-40, 16, 19, 10191322),
        (23, -53, -19, 10913505),
        (48, -20, 57, 10197540),
        (32, -72, -20, 10022494),
        (-35, 24, 27, 10402199),
        (-42, -26, 12, 10022496),
        (-33, 9, 29, 10191322),
        (2, 4, -2, 10349031),
        (22, 44, 32, 10349031),
        (-60, -36, 16, 10571235),
        (-53, -61, -29, 10913505),
        (-31, -18, 46, 10913505),
        (-40, 16, 19, 10191322),
        (14, -13, -10, 10197540),
        (-52, -28, 12, 10407201),
        (32, 61, -16, 10913505),
        (40, -16, 12, 10022496),
        (-58, -41, -20, 10913505),
        (26, -93, 2, 10227106),
        (-52, -16, 8, 10407201),
        (59, -19, -12, 10197540),
        (-29, -31, -2, 10197540),
        (-56, -43, 19, 10197540),
        (55, -3, -1, 10913505),
        (-49, 28, 4, 10227106),
        (-42, -26, 8, 10022496),
        (5, -39, 22, 10913505),
    )
    peaks = pd.DataFrame(peaks_data, columns=("x", "y", "z", "tfidf"))

    terms_data = (
        (10751455, "emotion", 0.052538),
        (10808134, "emotion", 0.244368),
        (10913505, "emotion", 0.059463),
        (11275483, "emotion", 0.123356),
        (11430815, "emotion", 0.065204),
        (9819274, "auditory", 0.059705),
        (10751455, "auditory", 0.052538),
        (10022494, "auditory", 0.10052),
        (10022496, "auditory", 0.28146),
        (10197540, "auditory", 0.38657),
        (10407201, "auditory", 0.16817),
        (11430815, "auditory", 0.065204),
        (10191322, "language", 0.28298),
        (10751455, "language", 0.052538),
        (10227106, "language", 0.05432),
        (10349031, "language", 0.05329),
        (10402199, "language", 0.18801),
        (10571235, "language", 0.08257),
        (11430815, "language", 0.065204),
    )
    terms = pd.DataFrame(terms_data, columns=("study", "term", "tfidf"))
    nl.add_tuple_set(peaks, name="PeakReported")
    nl.add_tuple_set(terms, name="TermInStudyTFIDF")
    nl.add_tuple_set(terms[["term"]].drop_duplicates(), name="Term")
    return nl


def test_neurosynth_query(nl):
    query = """
    TermInStudy(term, study) :: (1 / (1 + exp(-300 * (tfidf - 0.001)))) :- TermInStudyTFIDF(study, term, tfidf)
    TermAssociation(term) :- SelectedStudy(study) , TermInStudy(term, study)
    Activation(x, y, z) :- SelectedStudy(s) , PeakReported(x, y, z, s)
    ActivationGivenTerm(x, y, z, t, PROB) :- Activation(x, y, z) // TermAssociation(t)
    QueryActivation(x, y, z, p) :- ActivationGivenTerm(x, y, z, "emotion", p)
    ans(x, y, z, p) :- QueryActivation(x, y, z, p)
    """
    res = nl.execute_datalog_program(query)
    expected = (
        (-58, -41, -20, 0.2),
        (-53, -61, -29, 0.2),
        (-48, 32, -16, 0.2),
        (-31, -18, 46, 0.2),
        (0, -73, 6, 0.2),
        (5, -39, 22, 0.2),
        (23, -53, -19, 0.2),
        (32, 61, -16, 0.2),
        (38, -84, 0, 0.2),
        (55, -3, -1, 0.2),
    )
    assert res is not None
    assert len(res) == len(expected)
    p = res.as_pandas_dataframe()
    assert set(tuple(x) for x in p[["x", "y", "z"]].values) == set(
        (x, y, z) for x, y, z, _ in expected
    )



import gzip
import json
import logging
import os
import os.path
import sys
from concurrent.futures import Future
from io import BytesIO
from pathlib import Path
from typing import Optional
from uuid import uuid4

import matplotlib
import pandas as pd
import tornado.ioloop
import tornado.iostream
import tornado.options
import tornado.web
import tornado.websocket
import yaml
from neurolang.regions import ExplicitVBR, ExplicitVBROverlay
from tornado.options import define, options

from .engines import DestrieuxEngineConf, NeurosynthEngineConf
from .queries import NeurolangQueryManager
from .responses import (
    CustomQueryResultsEncoder,
    QueryResults,
    base64_encode_nifti,
)

define("port", default=8888, help="run on the given port", type=int)
define(
    "data_dir",
    default=str(Path.home() / "neurolang_data"),
    help="path of a directory where the downloaded datasets are stored",
    type=str,
)
define(
    "npm_build",
    default=False,
    help="force a build of the frontend app",
    type=bool,
)
static_path = str(Path(__file__).resolve().parent / "neurolang-web" / "dist")

LOG = logging.getLogger(__name__)


class Application(tornado.web.Application):
    """
    tornado Application. Defines the routes which the application listens on.
    Holds the NeurolangQueryManager which manages queries / results.

    Parameters
    ----------
    nqm : NeurolangQueryManager
        the query manager
    """

    def __init__(self, nqm: NeurolangQueryManager):
        self.nqm = nqm
        uuid_pattern = (
            r"[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}"
        )
        print(f"Serving static files from {static_path}")

        handlers = [
            (r"/v1/empty", EmptyHandler),
            (r"/v1/engines", EnginesHandler),
            (
                r"/v1/cancel/({uuid})".format(uuid=uuid_pattern),
                CancelHandler,
            ),
            (
                r"/v1/status/({uuid})".format(uuid=uuid_pattern),
                StatusHandler,
            ),
            (
                r"/v1/symbol/(.+)",
                SymbolsHandler,
            ),
            (
                r"/v1/statement",
                QueryHandler,
            ),
            (
                r"/v1/statementsocket",
                QuerySocketHandler,
            ),
            (
                r"/v1/atlas",
                NiftiiImageHandler,
            ),
            (
                r"/v1/download/(.+)",
                DownloadsHandler,
            ),
            (
                r"/v1/figure/({uuid})".format(uuid=uuid_pattern),
                MpltFigureHandler,
            ),
            (
                r"/(.*)",
                StaticFileOrDefaultHandler,
                {"path": static_path, "default_filename": "index.html"},
            ),
        ]
        dev = (
            os.environ.get("NEUROLANG_SERVER_MODE") is None
            or os.environ.get("NEUROLANG_SERVER_MODE") == "dev"
        )
        cookie_secret = (
            os.environ.get("NEUROLANG_COOKIE_SECRET")
            or "WWdY+9ILT/u/7hZ24ubLYDA5I1lfgEQMuwRVMp9PY4U="
        )
        settings = dict(
            cookie_secret=cookie_secret,
            template_path=os.path.join(os.path.dirname(__file__), "templates"),
            static_path=static_path,
            xsrf_cookies=not dev,
            debug=dev,
        )
        super().__init__(handlers, **settings)


class StaticFileOrDefaultHandler(tornado.web.StaticFileHandler):
    """
    When serving static files, if file does not exist on path, return
    index.html instead.

    This is only useful for dev mode, as static file handling is done
    by nginx in production.
    """
    def validate_absolute_path(self, root: str, absolute_path: str) -> Optional[str]:
        try:
            return super().validate_absolute_path(root, absolute_path)
        except tornado.web.HTTPError as e:
            if e.status_code == 404:
                return os.path.join(static_path, "index.html")
            raise e


class MainHandler(tornado.web.RequestHandler):
    """
    Serve the index page.
    """

    def get(self):
        self.render("index.html")


class EmptyHandler(tornado.web.RequestHandler):
    """
    Helper endpoint returning an empty
    result.
    """

    def get(self):
        return self.write({"id": str(uuid4())})


class CancelHandler(tornado.web.RequestHandler):
    """
    Cancel an already running computation.
    """

    def delete(self, uuid: str):
        LOG.debug("Canceling the request with uuid %s.", uuid)
        result = self.application.nqm.cancel(uuid)
        return self.write_json_reponse({"cancelled": result})


def query_results_to_json(data=None, status: str = "ok"):
    response = {"status": status}
    if data is not None:
        response["data"] = data
    return json.dumps(response, cls=CustomQueryResultsEncoder)


class JSONRequestHandler(tornado.web.RequestHandler):
    """
    Base Handler for writing JSON responses.
    """

    def set_default_headers(self):
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Headers", "x-requested-with")
        self.set_header("Access-Control-Allow-Methods", "POST, GET, OPTIONS")
        self.set_header("Content-Type", "application/json")

    def write_json_reponse(self, data=None, status: str = "ok"):
        return self.write(query_results_to_json(data, status))


class StatusHandler(JSONRequestHandler):
    async def get(self, uuid: str):
        """
        Return the status (or the result) of an already running calculation.
        Optional query parameters are :
            * symbol : return only the values for this symbol
            * start : index of the first row to return
            * length : number of rows to return
            * sort : the index of the column to sort by
            * asc : sort by ascending (true) or descending (false)

        Parameters
        ----------
        uuid : str
            The query id

        Returns
        -------
        QueryResults
            The query results.

        Raises
        ------
        tornado.web.HTTPError
            raises 404 if query id does not exist.
        """
        LOG.debug("Accessing status for request %s.", uuid)
        try:
            future = self.application.nqm.get_result(uuid)
        except KeyError:
            raise tornado.web.HTTPError(
                status_code=404, log_message="uuid not found"
            )
        symbol = self.get_argument("symbol", None)
        start = int(self.get_argument("start", 0))
        length = int(self.get_argument("length", 50))
        sort = int(self.get_argument("sort", -1))
        asc = bool(int(self.get_argument("asc", 1)))
        return self.write_json_reponse(
            QueryResults(uuid, future, symbol, start, length, sort, asc)
        )


class SymbolsHandler(JSONRequestHandler):
    def get(self, engine: str):
        """
        Return the symbols available on an engine.
        This method is syncronous and will block until an engine is
        available to get its symbols.

        Optional query parameters are :
            * symbol : return only the values for this symbol
            * start : index of the first row to return
            * length : number of rows to return
            * sort : the index of the column to sort by
            * asc : sort by ascending (true) or descending (false)

        Parameters
        ----------
        engine : str
            The engine id

        Returns
        -------
        QueryResults
            A dict of query results

        Raises
        ------
        tornado.web.HTTPError
            raises 404 if engine id does not exist
        """
        LOG.debug("Accessing symbols for engine %s.", engine)
        try:
            symbols = self.application.nqm.get_symbols(engine)
            # Block until results are available
            symbols.result()
        except KeyError:
            raise tornado.web.HTTPError(
                status_code=404, log_message="engine not found"
            )
        uuid = f"ENGINE_SYMBOLS_{engine}"
        symbol = self.get_argument("symbol", None)
        start = int(self.get_argument("start", 0))
        length = int(self.get_argument("length", 50))
        sort = int(self.get_argument("sort", -1))
        asc = bool(int(self.get_argument("asc", 1)))
        return self.write_json_reponse(
            QueryResults(uuid, symbols, symbol, start, length, sort, asc)
        )


class QuerySocketHandler(tornado.websocket.WebSocketHandler):
    """
    Main handler to submit a query using a websocket.
    """

    def check_origin(self, origin):
        """
        Allow cross-origin websockets.
        """
        return True

    async def on_message(self, message):
        """
        Upon receiving a message, extract the query from it and generate
        a uuid. Submit the query to the NeurolangQueryManager and set
        `send_query_update` as the callback for when the query execution
        completes.

        Parameters
        ----------
        message : str
            a serialized JSON representation of an object containing a query.
        """
        parsed = tornado.escape.json_decode(message)
        query = parsed["query"]
        engine = parsed.get("engine", "neurosynth")
        self.uuid = str(uuid4())

        LOG.debug("Submitting query with uuid %s.", self.uuid)
        future = self.application.nqm.submit_query(self.uuid, query, engine)
        tornado.ioloop.IOLoop.current().add_future(
            future, self.send_query_update
        )
        self.send_query_update(future)

    def send_query_update(self, future: Future, status: str = "ok"):
        """
        Upon completion of a query, send a message with the results

        Parameters
        ----------
        future : Future
            the query results.
        status : str, optional
            the status of the query, by default "ok"
        """
        self.write_message(
            query_results_to_json(QueryResults(self.uuid, future), status)
        )


class QueryHandler(JSONRequestHandler):
    """
    Main endpoint to submit a query using a POST request.
    """

    async def post(self):
        query = self.get_argument("query")
        engine = self.get_argument("engine", "neurosynth")
        uuid = str(uuid4())
        LOG.debug("Submitting query with uuid %s.", uuid)
        self.application.nqm.submit_query(uuid, query, engine)
        return self.write_json_reponse({"query": query, "uuid": uuid})


class NiftiiImageHandler(JSONRequestHandler):
    """
    Return the atlas image to be used by the Papaya viewer.
    Currently returns the MNI_MASK used by the engine configuration.

    Image is returned as base64 encoded.
    """

    def get(self):
        engine = self.get_argument("engine", "neurosynth")
        atlas = self.application.nqm.get_atlas(engine)
        return self.write_json_reponse({"image": base64_encode_nifti(atlas)})


class DownloadsHandler(tornado.web.RequestHandler):
    """
    Handle requests for file downloads.
    """

    def set_default_headers(self):
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Headers", "x-requested-with")
        self.set_header("Access-Control-Allow-Methods", "POST, GET, OPTIONS")
        self.set_header("Content-Type", "application/octet-stream")

    def save_image_to_gzip(
        self, df: pd.DataFrame, col: str, idx: str
    ) -> BytesIO:
        """
        Save an image as nii.gz to a bytes array. The image should be an
        ExplicitVBR or ExplicitVBROverlay object located in the given
        dataframe at idx, col position.

        Parameters
        ----------
        df : pd.DataFrame
            the dataframe containing the images
        col : str
            the col index for the image in the df
        idx : str
            the row index for the image in the df

        Returns
        -------
        BytesIO
            the bytes buffer with the compressed image data

        Raises
        ------
        tornado.web.HTTPError
            404 if object at given col, row indices is not the right type.
        """
        image = df.iat[int(idx), int(col)]
        if not isinstance(image, (ExplicitVBR, ExplicitVBROverlay)):
            raise tornado.web.HTTPError(
                status_code=404, log_message="Invalid file format to download"
            )
        data = BytesIO(gzip.compress(image.spatial_image().dataobj.tobytes()))
        return data

    async def get(self, key: str):
        """
        Serve the file for the given key. Required query parameters are:
        - symbol: the symbol to download
        Optional query parameters:
        - col
        - idx
        the col and row index when a specific image file should be downloaded
        instead of the whole dataframe.

        Parameters
        ----------
        key : str
            the unique id key, either for a query or for an engine type

        Raises
        ------
        tornado.web.HTTPError
            404 if key is invalid, or results are not available
        """
        # 1. Get the symbol to download
        symbol = self.get_argument("symbol")
        col = self.get_argument("col", None)
        idx = self.get_argument("idx", None)
        try:
            future = self.application.nqm.get_result(key)
            LOG.debug(
                "Preparing image file for query %s and symbol %s.", key, symbol
            )
        except KeyError:
            try:
                future = self.application.nqm.get_symbols(key)
                LOG.debug(
                    f"Preparing image file for engine %s and symbol %s.",
                    key,
                    symbol,
                )
            except KeyError:
                raise tornado.web.HTTPError(
                    status_code=404, log_message="uuid not found"
                )
        if not (future.done() and future.result()):
            raise tornado.web.HTTPError(
                status_code=404, log_message="query does not have results"
            )
        results = future.result()
        ras = results[symbol]
        df = ras.as_pandas_dataframe()

        # 2. Write the object as gzip bytes
        if col is not None and idx is not None:
            data = self.save_image_to_gzip(df, col, idx)
            filename = symbol + "_" + idx + ".nii.gz"
        else:
            data = BytesIO(
                gzip.compress(
                    df.to_csv(index=False, compression="gzip").encode()
                )
            )
            filename = symbol + ".csv.gz"

        # 3. Stream the data in chunks to avoid blocking the server
        self.set_header(
            "Content-Disposition", "attachment; filename=" + filename
        )
        chunk_size = 1024 * 1024 * 1  # 1 MiB
        while True:
            chunk = data.read(chunk_size)
            if not chunk:
                break
            try:
                self.write(chunk)  # write the chunk to response
                await self.flush()  # send the chunk to client
            except tornado.iostream.StreamClosedError:
                break
            finally:
                del chunk


class MpltFigureHandler(tornado.web.RequestHandler):
    """
    Handle requests for matplotlib figures. Streams them as svg.
    """

    def set_default_headers(self):
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Headers", "x-requested-with")
        self.set_header("Access-Control-Allow-Methods", "POST, GET, OPTIONS")
        self.set_header("Content-Type", "image/svg+xml")

    async def get(self, uuid: str):
        """
        Serve the matplotlib figure requested. Required query parameters are:
        - symbol: the symbol containing the figures
        - col & row: the col and row index of the figure in the symbol's dataframe.

        Parameters
        ----------
        uuid : str
            the uuid of the query

        Raises
        ------
        tornado.web.HTTPError
            404 if uuid is invalid, or results are not available
        """
        # 1. Get the result dataframe for the query params
        symbol = self.get_argument("symbol")
        row = self.get_argument("row")
        col = self.get_argument("col")
        format_ext = self.get_argument("format", "svg")
        LOG.debug("Accessing figure result for request %s.", uuid)
        try:
            future = self.application.nqm.get_result(uuid)
        except KeyError:
            raise tornado.web.HTTPError(
                status_code=404, log_message="uuid not found"
            )
        if not (future.done() and future.result()):
            raise tornado.web.HTTPError(
                status_code=404, log_message="query does not have results"
            )
        results = future.result()
        ras = results[symbol]
        df = ras.as_pandas_dataframe()
        figure = df.iat[int(row), int(col)]
        if not isinstance(figure, matplotlib.figure.Figure):
            raise tornado.web.HTTPError(
                status_code=404, log_message="Invalid figure format"
            )

        # 2. Stream the figure in requested format. If format is not svg
        # we serve it as an attached file
        if format_ext != "svg":
            filename = f"{symbol}_{row}_{col}.{format_ext}"
            self.set_header(
                "Content-Disposition", "attachment; filename=" + filename
            )
        data = BytesIO()
        figure.savefig(data, format=format_ext)
        data.seek(0)
        chunk_size = 1024 * 1024 * 1  # 1 MiB
        while True:
            chunk = data.read(chunk_size)
            if not chunk:
                break
            try:
                self.write(chunk)  # write the chunk to response
                await self.flush()  # send the chunk to client
            except tornado.iostream.StreamClosedError:
                break
            finally:
                del chunk


class EnginesHandler(JSONRequestHandler):
    def get(self):
        engines = [e.key for e in self.application.nqm.configs.keys()]

        dirs = [
            os.path.join(sys.prefix, "queries"),
            os.path.dirname(os.path.realpath(__file__)),
        ]
        for d in dirs:
            queries_file = os.path.join(d, "queries.yaml")
            if os.path.isfile(queries_file):
                LOG.info(
                    "Reading queries configuration file for Neurolang: %s",
                    queries_file,
                )
                with open(queries_file, "r") as stream:
                    queries = yaml.safe_load(stream)
                break

        data = []
        for engine in engines:
            res = {"engine": engine}
            if engine in queries:
                res["queries"] = queries[engine]
            data.append(res)
        self.write_json_reponse(data)


def setup_logs():
    logger = logging.getLogger("neurolang.utils.server")
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    logger.propagate = False

    logger = logging.getLogger("neurolang")
    logger.propagate = False
    LOG.setLevel(logging.DEBUG)


def main():
    tornado.options.parse_command_line()
    setup_logs()
    data_dir = Path(options.data_dir)
    LOG.info(f"Neurolang data directory set to {data_dir}")
    opts = {
        NeurosynthEngineConf(data_dir, resolution=2): 2,
        DestrieuxEngineConf(data_dir): 2,
    }
    nqm = NeurolangQueryManager(opts)

    print(
        f"Tornado application starting on http://localhost:{options.port}/ ..."
    )
    app = Application(nqm)
    app.listen(options.port)
    tornado.ioloop.IOLoop.current().start()


if __name__ == "__main__":
    main()



import logging
import re
import types
import uuid
from typing import Iterable, Tuple, Union

import dask.array as da
import dask.dataframe as dd
import numpy as np
import pandas as pd
from sqlalchemy import (
    and_,
    column,
    func,
    literal,
    literal_column,
    select,
    text
)
from sqlalchemy.sql import FromClause, except_, intersect, table, union
from typing_extensions import get_origin

from ...config import config
from ...type_system import Unknown, get_args, infer_type, is_parameterized
from . import abstract as abc
from .dask_helpers import (
    DaskContextManager,
    convert_type_to_pandas_dtype,
    convert_types_to_pandas_dtype,
    timeit,
    try_to_infer_type_of_operation
)

LOG = logging.getLogger(__name__)


NA = None


class RelationalAlgebraStringExpression(str):
    """
    RelationalAlgebraStringExpression which replaces some pandas/python
    operators by SQL compatible ones.
    """

    def __new__(cls, content):
        instance = super().__new__(cls, content)
        instance = super().__new__(cls, cls._match_operations(str(instance)))
        return instance

    @classmethod
    def _match_operations(cls, content):
        clean_content = cls._match_not_equal(content)
        clean_content = cls._match_power(clean_content)
        clean_content = cls._match_log(clean_content)
        return clean_content

    @staticmethod
    def _match_not_equal(content):
        """
        Replace != by <>
        """
        return re.sub("!=", "<>", content)

    @staticmethod
    def _match_power(content):
        """
        Replace (d ** 2) by POWER(d, 2)
        """
        p = re.compile(r"\(?([^\*]+)\*\*([^\*\)]+)\)?")
        m = p.match(content)
        if m:
            return f"POWER({m.group(1)}, {m.group(2)})"
        return content

    @staticmethod
    def _match_log(content):
        """
        Replace log(x) by ln(x)
        """
        p = re.compile(r"log\((.+)\)")
        m = p.match(content)
        if m:
            return f"ln({m.group(1)})"
        return content

    def __repr__(self):
        return "{}{{ {} }}".format(self.__class__.__name__, super().__repr__())


def _new_name(prefix="table_"):
    if config["RAS"].get("tableIds", "uuid") == "uuid":
        return prefix + str(uuid.uuid4()).replace("-", "_")
    else:
        return prefix + next(DaskContextManager._id_generator_)


class DaskRelationalAlgebraBaseSet:
    """
    Base class for RelationalAlgebraSets relying on a Dask-SQL backend.
    This class defines no RA operations but has all the logic of creating /
    iterating / fetching of items in the sets.
    """

    _count = None
    _is_empty = None
    _table_name = None
    _table = None
    _container = None
    _init_columns = None
    row_types = None

    def __init__(self, iterable=None, columns=None):
        self._init_columns = columns
        if isinstance(iterable, DaskRelationalAlgebraBaseSet):
            if (
                columns is None
                or len(columns) == 0
                or columns == iterable.columns
            ):
                self._init_from(iterable)
            else:
                self._init_from_and_rename(iterable, columns)
        elif iterable is not None:
            self._create_insert_table(iterable, columns)

    def _init_from(self, other):
        self._table_name = other._table_name
        self._container = other._container
        self._table = other._table
        self._count = other._count
        self._is_empty = other._is_empty
        self._init_columns = other._init_columns
        self.row_types = other.row_types

    def _init_from_and_rename(self, other, columns):
        if other._table is not None:
            query = select(
                *[
                    c.label(str(nc))
                    for c, nc in zip(other.sql_columns, columns)
                ]
            ).select_from(other._table)
            self._table = query.subquery()
        self._is_empty = other._is_empty
        self._count = other._count
        if other.row_types is not None:
            self.row_types = other.row_types.rename(
                {other.row_types.index[i]: c for i, c in enumerate(columns)}
            )

    def _create_insert_table(self, data, columns=None):
        if not isinstance(data, pd.DataFrame):
            data = pd.DataFrame(data, columns=columns)
        elif columns is not None:
            data.columns = list(columns)
        data.columns = data.columns.astype(str)
        data = data.drop_duplicates()
        if len(data) > 0 and len(data.columns) > 0:
            # partitions should not be too small, yet fit nicely in memory.
            # The amount of RAM available on the machine should be greater
            # than nb of core x partition size.
            # Here we create partitions of less than 500mb based on the
            # original df size.
            df_size = data.memory_usage(deep=True).sum() / (1 << 20)
            npartitions = 1 + int(df_size) // 500
            LOG.info(
                "Creating dask dataframe with {} partitions"
                " from {:0.2f} Mb pandas df.".format(npartitions, df_size)
            )
            ddf = dd.from_pandas(data, npartitions=npartitions)
            self._set_container(ddf, persist=True)
            self.row_types = pd.Series(
                get_args(infer_type(next(data.itertuples(index=False)))),
                index=data.columns,
            )
        self._count = len(data)
        self._is_empty = self._count == 0

    @timeit
    def _set_container(self, ddf, persist=True, prefix="table_", sql=None):
        self._container = ddf
        if persist:
            # Persist triggers an evaluation of the dask dataframe task-graph.
            # This evaluation is asynchronous (if using an asynchronous scheduler).
            # It will return a new dataframe with a shallow graph.
            # See https://distributed.dask.org/en/latest/memory.html#persisting-collections
            # Since we're calling persist here, make sure to pass persist=False to
            # create_table method to not call it twice.
            self._container = self._container.persist()
            self._table_name = _new_name(prefix=prefix)
            DaskContextManager.get_context().create_table(
                self._table_name, ddf, persist=False, sql=sql
            )
            self._table = table(
                self._table_name, *[column(c) for c in ddf.columns]
            )

    @property
    def set_row_type(self):
        """
        Return typing info for this set.
        """
        if self.arity > 0:
            if self.row_types is not None:
                return Tuple[tuple(self.row_types.values)]
            return Tuple[tuple(Unknown for _ in self.columns)]
        return Tuple

    @property
    def dtypes(self, default_type=np.float64):
        """
        return a pandas dtypes array containing numpy dtypes for each column.
        """
        if self.row_types is None:
            return None
        return convert_types_to_pandas_dtype(self.row_types, default_type)

    @property
    def arity(self):
        return len(self.columns)

    @property
    def columns(self):
        if self._table is None:
            return [] if self._init_columns is None else self._init_columns
        return self._table.c.keys()

    @property
    def sql_columns(self):
        if self._table is None:
            return {}
        return self._table.c

    @property
    def container(self):
        """
        Accessing the container will evaluate the SQL query representing this set and
        persist the results in Dask.
        """
        if self._container is None:
            if self._table is not None and self.arity > 0:
                q = select(self._table)
                ddf = DaskContextManager.sql(q)
                ddf.columns = self.columns
                query = DaskContextManager.compile_query(q)
                self._set_container(
                    ddf, persist=True, prefix="table_as_", sql=query
                )
        return self._container

    @classmethod
    def dee(cls):
        output = cls()
        output._count = 1
        output._is_empty = False
        return output

    @classmethod
    def dum(cls):
        output = cls()
        output._count = 0
        return output

    @classmethod
    def create_view_from(cls, other):
        if not isinstance(other, cls):
            raise ValueError(
                "View can only be created from an object of the same class"
            )
        output = cls()
        output._init_from(other)
        return output

    def copy(self):
        if self.is_dee():
            return self.dee()
        elif self.is_dum():
            return self.dum()
        return type(self).create_view_from(self)

    def _create_view_from_query(
        self,
        query: FromClause,
        row_types: pd.Series = None,
        is_empty: bool = None,
    ):
        """
        This method is called to create a new set whenever a relational
        algebra operation is performed on this set.

        Parameters
        ----------
        query : FromClause
            The sql query for the new set
        row_types : pd.Series, optional
            typings row type information for the newset,
            if None self.row_types will be used
        is_empty : bool, optional
            is the new set empty, by default None

        Returns
        -------
        DaskRelationalAlgebraBaseSet
            A new set corresponding to the query
        """
        output = type(self)()
        output._table = query.subquery()
        output._container = None
        output._is_empty = is_empty
        if row_types is None:
            row_types = self.row_types
        output.row_types = row_types
        return output

    def is_empty(self):
        if self._is_empty is None:
            if self._count is not None:
                self._is_empty = self._count == 0
            else:
                self._is_empty = self.fetch_one() is None
        return self._is_empty

    def is_dum(self):
        return self.arity == 0 and self.is_empty()

    def is_dee(self):
        return self.arity == 0 and not self.is_empty()

    def __len__(self):
        if self._count is None:
            if self.container is None:
                self._count = 0
            else:
                self._count = len(self.container.drop_duplicates())
        return self._count

    def __contains__(self, element):
        if self.arity == 0:
            return False
        element = self._normalise_element(element)
        query = select(self._table)
        for c, v in element.items():
            query = query.where(self.sql_columns.get(c) == v)
        res = DaskContextManager.sql(query).head(1, npartitions=-1)
        return len(res) > 0

    def _normalise_element(self, element):
        if isinstance(element, dict):
            pass
        elif hasattr(element, "__iter__") and not isinstance(element, str):
            element = dict(zip(self.columns, element))
        else:
            element = dict(zip(self.columns, (element,)))
        return element

    def itervalues(self):
        if self.is_dee():
            return iter([tuple()])
        else:
            return iter(self._fetchall().itertuples(name=None, index=False))

    def __iter__(self):
        return self._do__iter__(named=False)

    def _do__iter__(self, named=False):
        if self.is_dee():
            return iter([tuple()])
        if named:
            try:
                return self._fetchall().itertuples(name="tuple", index=False)
            except ValueError:
                # Invalid column names for namedtuple, just return unnamed tuples
                pass
        return self._fetchall().itertuples(name=None, index=False)

    def as_numpy_array(self):
        return self._fetchall().to_numpy()

    def as_pandas_dataframe(self):
        df = self._fetchall()
        try:
            df.columns = df.columns.astype(int)
        except TypeError:
            pass
        return df

    def _fetchall(self):
        if self.container is None:
            if self._count == 1:
                return pd.DataFrame([()])
            else:
                return pd.DataFrame([])
        df = self.container.compute()
        return df

    def fetch_one(self):
        return self._fetch_one(named=False)

    def _fetch_one(self, named=False):
        if self.container is None:
            if self._count == 1:
                return tuple()
            return None
        if not hasattr(self, "_one_row"):
            name = "tuple" if named else None
            try:
                self._one_row = next(
                    self.container.head(1).itertuples(name=name, index=False)
                )
            except StopIteration:
                self._one_row = None
        return self._one_row

    def __eq__(self, other):
        if isinstance(other, DaskRelationalAlgebraBaseSet):
            if self.is_dee() or other.is_dee():
                res = self.is_dee() and other.is_dee()
            elif self.is_dum() or other.is_dum():
                res = self.is_dum() and other.is_dum()
            elif (
                self._table_name is not None
                and self._table_name == other._table_name
            ):
                res = True
            elif not self._equal_sets_structure(other):
                res = False
            else:
                scont = self.container
                ocont = other.container
                if (
                    scont is None
                    or len(scont) == 0
                    or ocont is None
                    or len(ocont) == 0
                ):
                    res = (scont is None or len(scont) == 0) and (
                        ocont is None or len(ocont) == 0
                    )
                else:
                    intersection_dups = (
                        scont.merge(ocont, how="outer", indicator=True)
                        .iloc[:, -1]
                        .compute()
                    )
                    res = (intersection_dups == "both").all()
            return res
        else:
            return super().__eq__(other)

    def _equal_sets_structure(self, other):
        if isinstance(self, abc.NamedRelationalAlgebraFrozenSet) or isinstance(
            other, abc.NamedRelationalAlgebraFrozenSet
        ):
            return set(self.columns) == set(other.columns)
        return self.arity == other.arity

    def __repr__(self):
        t = self._table
        return "{}({})".format(type(self), t)

    def __hash__(self):
        if self._table is None:
            return hash((tuple(), None))
        return hash((tuple(self.columns), self.as_numpy_array().tobytes()))


class RelationalAlgebraFrozenSet(
    DaskRelationalAlgebraBaseSet, abc.RelationalAlgebraFrozenSet
):
    def __init__(self, iterable=None, columns=None):
        super().__init__(iterable, columns=columns)

    def selection(self, select_criteria):
        if self._table is None:
            return self.copy()

        query = select(self._table)
        if callable(select_criteria):
            lambda_name = _new_name("lambda")
            params = [(c, self.dtypes[c]) for c in self.columns]
            DaskContextManager.register_function(
                select_criteria, lambda_name, params, np.bool8, True
            )
            f_ = getattr(func, lambda_name)
            query = query.where(f_(*self.sql_columns))
        elif isinstance(select_criteria, RelationalAlgebraStringExpression):
            # replace == used in python by = used in SQL
            query = query.where(text(re.sub("==", "=", str(select_criteria))))
        else:
            for k, v in select_criteria.items():
                if callable(v):
                    lambda_name = _new_name("lambda")
                    c_ = self.sql_columns.get(str(k))
                    DaskContextManager.register_function(
                        v,
                        lambda_name,
                        [(str(k), self.dtypes[str(k)])],
                        np.bool8,
                        False,
                    )
                    f_ = getattr(func, lambda_name)
                    query = query.where(f_(c_))
                elif isinstance(
                    select_criteria, RelationalAlgebraStringExpression
                ):
                    query = query.where(text(re.sub("==", "=", str(v))))
                else:
                    query = query.where(self.sql_columns.get(str(k)) == v)
        return self._create_view_from_query(query)

    def selection_columns(self, select_criteria):
        if self._table is None:
            return self.copy()
        query = select(*self.sql_columns).select_from(self._table)
        for k, v in select_criteria.items():
            query = query.where(
                self.sql_columns.get(str(k)) == self.sql_columns.get(str(v))
            )
        return self._create_view_from_query(query)

    def equijoin(self, other, join_indices):
        res = self._dee_dum_product(other)
        if res is not None:
            return res

        # Create an alias on the other table's name if we're joining on
        # the same table.
        ot = other._table
        if other._table_name == self._table_name:
            ot = ot.alias()

        join_cols = list(self.sql_columns) + [
            ot.c.get(str(i)).label(str(i + self.arity))
            for i in range(other.arity)
        ]
        query = select(*join_cols)

        if join_indices is not None and len(join_indices) > 0:
            on_clause = and_(
                *[
                    self.sql_columns.get(str(i)) == ot.c.get(str(j))
                    for i, j in join_indices
                ]
            )
            query = query.select_from(self._table.join(ot, on_clause))
        row_types = pd.concat(
            [
                self.row_types,
                other.row_types.rename(
                    {str(i): str(i + self.arity) for i in range(other.arity)}
                ),
            ]
        )
        return self._create_view_from_query(query, row_types=row_types)

    def cross_product(self, other):
        return self.equijoin(other, join_indices=None)

    def groupby(self, columns):
        if self.container is not None:
            if isinstance(columns, str) or not isinstance(columns, Iterable):
                columns = [columns]
            columns = list(map(str, columns))
            df = self.container.compute()
            for g_id, group in df.groupby(by=columns, dropna=False):
                group_set = type(self)(iterable=group)
                yield g_id, group_set

    def projection(self, *columns):
        return self._do_projection(*columns, reindex=True)

    def _do_projection(self, *columns, reindex=True):
        if len(columns) == 0 or self.arity == 0:
            new = type(self)()
            if len(self) > 0:
                new._count = 1
            return new

        if self._table is None:
            return type(self)(columns=columns, iterable=[])

        row_types = self.row_types[[str(c) for c in columns]]
        if reindex:
            proj_columns = [
                self.sql_columns.get(str(c)).label(str(i))
                for i, c in enumerate(columns)
            ]
            row_types.index = [str(i) for i in range(len(columns))]
        else:
            proj_columns = [self.sql_columns.get(str(c)) for c in columns]
        query = select(proj_columns).select_from(self._table)
        if len(set(proj_columns)) != len(set(self.sql_columns)):
            query = query.distinct()

        return self._create_view_from_query(
            query, row_types=row_types, is_empty=self._is_empty
        )

    def _do_set_operation(self, other, sql_operator):
        if not self._equal_sets_structure(other):
            raise ValueError(
                "Relational algebra set operators can only be used on sets"
                " with same columns."
            )

        ot = other._table
        if other._table_name == self._table_name:
            ot = ot.alias()
        query = sql_operator(
            select(self._table),
            select([ot.c.get(c) for c in self.columns]).select_from(ot),
        )
        if sql_operator is union:
            query = select(query.subquery()).distinct()
        return self._create_view_from_query(query)

    def __and__(self, other):
        if not isinstance(other, RelationalAlgebraFrozenSet):
            return super().__and__(other)
        res = self._dee_dum_product(other)
        if res is not None:
            return res
        if self._table is None or other._table is None:
            return type(self)(columns=self.columns, iterable=[])
        return self._do_set_operation(other, intersect)

    def __or__(self, other):
        if not isinstance(other, RelationalAlgebraFrozenSet):
            return super().__or__(other)
        res = self._dee_dum_sum(other)
        if res is not None:
            return res
        if self._table is None:
            return other.copy()
        elif other._table is None:
            return self.copy()
        return self._do_set_operation(other, union)

    def __sub__(self, other):
        if not isinstance(other, RelationalAlgebraFrozenSet):
            return super().__sub__(other)
        if self.arity == 0 or other.arity == 0:
            if self.is_dee() and other.is_dee():
                return self.dum()
            return self.copy()
        if self._table is None or other._table is None:
            return self.copy()
        return self._do_set_operation(other, except_)


class NamedRelationalAlgebraFrozenSet(
    RelationalAlgebraFrozenSet, abc.NamedRelationalAlgebraFrozenSet
):
    def __init__(self, columns=None, iterable=None):
        if isinstance(columns, RelationalAlgebraFrozenSet):
            iterable = columns
            columns = columns.columns
        self._check_for_duplicated_columns(columns)
        super().__init__(iterable, columns)

    @staticmethod
    def _check_for_duplicated_columns(columns):
        if columns is not None and len(set(columns)) != len(columns):
            columns = list(columns)
            dup_cols = set(c for c in columns if columns.count(c) > 1)
            raise ValueError(
                "Duplicated column names are not allowed. "
                f"Found the following duplicated columns: {dup_cols}"
            )

    @property
    def columns(self):
        return tuple(super().columns)

    def fetch_one(self):
        return super()._fetch_one(named=True)

    def __iter__(self):
        return super()._do__iter__(named=True)

    def projection(self, *columns):
        return super()._do_projection(*columns, reindex=False)

    def cross_product(self, other):
        res = self._dee_dum_product(other)
        if res is not None:
            return res
        if len(set(self.columns).intersection(set(other.columns))) > 0:
            raise ValueError(
                "Cross product with common columns " "is not valid"
            )

        if self._table is None or other._table is None:
            return self._do_empty_join(other, isouter=False)

        query = select(self._table, other._table)
        return self._create_view_from_query(
            query,
            row_types=self.row_types.append(other.row_types),
            is_empty=self._is_empty,
        )

    def naturaljoin(self, other):
        res = self._dee_dum_product(other)
        if res is not None:
            return res

        on = [c for c in self.columns if c in other.columns]
        if len(on) == 0:
            return self.cross_product(other)

        if self._table is None or other._table is None:
            return self._do_empty_join(other, isouter=False)
        return self._do_join(other, on, isouter=False)

    def left_naturaljoin(self, other):
        """
        Same as naturaljoin with outer=True
        """
        if self.is_dee():
            return self
        on = [c for c in self.columns if c in other.columns]
        if len(on) == 0:
            if other.is_empty():
                columns = {
                    c: RelationalAlgebraStringExpression(c)
                    for c in self.columns
                }
                for c in other.columns:
                    columns[c] = RelationalAlgebraStringExpression("NULL")
                return self.extended_projection(columns)
            else:
                return self.cross_product(other)
        if self._table is None or other._table is None:
            return self._do_empty_join(other, isouter=True)
        return self._do_join(other, on, isouter=True)

    def _do_join(self, other, on, isouter=False):
        """
        Performs the join on the two sets.

        Parameters
        ----------
        other : NamedRelationalAlgebraFrozenSet
            The other set
        on : Iterable[sqlalchemy.Columns]
            The columns to join on
        isouter : bool, optional
            If True, performs a left outer join, by default False

        Returns
        -------
        NamedRelationalAlgebraFrozenSet
            The joined set
        """
        # Create an alias on the other table's name if we're joining on
        # the same table.
        ot = other._table
        if other._table_name == self._table_name:
            ot = ot.alias()

        on_clause = and_(
            *[self._table.c.get(col) == ot.c.get(col) for col in on]
        )
        other_cols = list(set(other.columns) - set(self.columns))
        select_cols = [self._table] + [ot.c.get(col) for col in other_cols]
        query = select(*select_cols).select_from(
            self._table.join(ot, on_clause, isouter=isouter)
        )
        row_types = self.row_types.append(other.row_types[other_cols])
        empty = self._is_empty if isouter else None
        return self._create_view_from_query(
            query, row_types=row_types, is_empty=empty
        )

    def _do_empty_join(self, other, isouter=False):
        """
        Perform a join when self._table or other._table is None.
        Result will be an empty set with combined columns unless
        it's a left join and self._table is not None.
        """
        other_cols = tuple(set(other.columns) - set(self.columns))
        new_cols = self.columns + other_cols
        if not isouter or self._table is None:
            return type(self)(columns=new_cols, iterable=[])

        projections = {
            col: abc.RelationalAlgebraColumnStr(col) for col in self.columns
        }
        for col in other_cols:
            projections[col] = None
        return self.extended_projection(projections)

    def replace_null(self, dst_column, value):
        columns = self.columns
        if len(columns) == 0 or self.arity == 0:
            new = type(self)()
            if len(self) > 0:
                new._count = 1
            return new

        if self._table is None:
            return type(self)(columns=columns, iterable=[])

        sql_dst_column = self.sql_columns.get(dst_column)
        columns_without_change = [
            c for c in self.sql_columns if c != sql_dst_column
        ]
        if isinstance(value, RelationalAlgebraStringExpression):
            if str(value) != str(dst_column):
                value = literal_column(value)
            else:
                value = self.sql_columns.get(str(value))
        elif isinstance(value, abc.RelationalAlgebraColumn):
            value = self.sql_columns.get(str(value))
        else:
            value = literal(value)

        proj_column = func.coalesce(sql_dst_column, value).label(dst_column)
        query = (
            select(
                columns_without_change +
                [proj_column]
            )
            .select_from(self._table)
        )

        return self._create_view_from_query(
            query, row_types=self.row_types
        )

    def explode(self, src_column: str, dst_columns: Union[str, Tuple[str]]):
        """
        Transform each element of a list-like column to a row.

        Since explode is not a standard SQL statement but is an operation
        implemented on dask dataframes, this relational algebra operation
        evaluates the dask container for the set on which it is applied, and
        then calls the `explode` method on it.

        Parameters
        ----------
        src_column : str
            The column to explode
        dst_columns: Union[str, Tuple[str]]
            The destination columns

        Returns
        -------
        NamedRelationalAlgebraFrozenSet
            The set with exploded column

        Examples
        --------
        >>> ras = NamedRelationalAlgebraFrozenSet(
        ...        columns=("x", "y", "z"),
        ...        iterable=[
        ...            (5, frozenset({1, 2, 5, 6}), "foo"),
        ...            (10, frozenset({5, 9}), "bar"),
        ...        ])
        >>> ras.explode('y')
            x  y    z
        0   5  1  foo
        1   5  2  foo
        2   5  5  foo
        3   5  6  foo
        4  10  9  bar
        5  10  5  bar
        """
        if self.is_dum() or self.is_dee():
            return self

        if self._table is None:
            return type(self)(columns=self.columns, iterable=[])

        q = select(self._table)
        ddf = DaskContextManager.sql(q)

        if dst_columns == src_column:
            # 1. replace original column by exploded one
            new_ddf = ddf.explode(src_column)
            new_columns = self.columns
            col_type = self.row_types[src_column]
        elif not isinstance(dst_columns, tuple):
            # 2. add new column with exploded values
            new_ddf = ddf.assign(**{dst_columns: ddf[src_column]}).explode(
                dst_columns
            )
            new_columns = self.columns + (dst_columns,)
            col_type = self.row_types[src_column]
        else:
            # 3. explode values into multiple columns
            new_ddf = ddf.assign(**{dst_columns[-1]: ddf[src_column]}).explode(
                dst_columns[-1]
            )
            for c, v in zip(dst_columns, zip(*new_ddf[dst_columns[-1]])):
                new_ddf[c] = da.from_array(v)
            new_columns = self.columns + dst_columns
            new_ddf = new_ddf[list(new_columns)]
            col_type = None

        if (
            is_parameterized(col_type)
            and issubclass(get_origin(col_type), Iterable)
            and len(get_args(col_type)) == 1
        ):
            col_type = get_args(col_type)[0]
            if col_type != Unknown:
                try:
                    new_ddf[dst_columns] = new_ddf[dst_columns].astype(
                        col_type
                    )
                except TypeError:
                    LOG.warn(
                        f"Unable to cast exploded column to type {col_type}"
                    )
        output = type(self)(columns=new_columns, iterable=new_ddf.compute())
        return output

    def equijoin(self, other, join_indices):
        raise NotImplementedError()

    def rename_column(self, src, dst):
        if src == dst:
            return self
        if dst in self.columns:
            raise ValueError(
                "Duplicated column names are not allowed. "
                f"{dst} is already a column name."
            )
        if self._table is None:
            new_cols = [dst if c == src else c for c in self.columns]
            return type(self)(columns=new_cols, iterable=[])
        query = select(
            *[
                c.label(str(dst)) if c.name == src else c
                for c in self.sql_columns
            ]
        ).select_from(self._table)
        row_types = self.row_types.rename({str(src): str(dst)})
        return self._create_view_from_query(
            query, row_types=row_types, is_empty=self._is_empty
        )

    def rename_columns(self, renames):
        # prevent duplicated destination columns
        self._check_for_duplicated_columns(renames.values())

        if self.is_dum():
            return self

        if not set(renames).issubset(self.columns):
            # get the missing source columns
            # for a more convenient error message
            not_found_cols = set(c for c in renames if c not in self.columns)
            raise ValueError(
                f"Cannot rename non-existing columns: {not_found_cols}"
            )

        if self._table is None:
            new_cols = [
                renames.get(c) if c in renames else c for c in self.columns
            ]
            return type(self)(columns=new_cols, iterable=[])

        query = select(
            *[
                c.label(str(renames.get(c.name))) if c.name in renames else c
                for c in self.sql_columns
            ]
        ).select_from(self._table)
        row_types = self.row_types.rename(renames)
        return self._create_view_from_query(
            query, row_types=row_types, is_empty=self._is_empty
        )

    def aggregate(self, group_columns, aggregate_function):
        group_columns = list(group_columns)
        if len(set(group_columns)) < len(group_columns):
            raise ValueError("Cannot group on repeated columns")
        if self.is_dee():
            raise ValueError(
                "Aggregation on non-empty sets with arity == 0 is unsupported."
            )
        distinct_sub_query = select(self._table).subquery()
        agg_cols, agg_types = self._build_aggregate_functions(
            group_columns, aggregate_function, distinct_sub_query
        )

        if self._table is None:
            new_cols = group_columns + list(col.name for col in agg_cols)
            return type(self)(columns=new_cols, iterable=[])

        groupby = [distinct_sub_query.c.get(str(c)) for c in group_columns]
        query = select(*(groupby + agg_cols)).group_by(*groupby)
        row_types = self.row_types[list(group_columns)].append(agg_types)
        return self._create_view_from_query(
            query, row_types=row_types, is_empty=self._is_empty
        )

    def _build_aggregate_functions(
        self, group_columns, aggregate_function, distinct_view
    ):
        """
        Create the list of aggregated destination columns.
        """
        if isinstance(aggregate_function, dict):
            agg_iter = ((k, k, v) for k, v in aggregate_function.items())
        elif isinstance(aggregate_function, (tuple, list)):
            agg_iter = aggregate_function
        else:
            raise ValueError(
                "Unsupported aggregate_function: {} of type {}".format(
                    aggregate_function, type(aggregate_function)
                )
            )
        un_grouped_cols = [
            c_ for c_ in distinct_view.c if c_.name not in group_columns
        ]
        agg_cols = []
        agg_types = pd.Series(dtype="object")
        for dst, src, f in agg_iter:
            if src in distinct_view.c.keys():
                # call the aggregate function on only one column
                c_ = [distinct_view.c.get(src)]
            else:
                # call the aggregate function on all the non-grouped columns
                c_ = un_grouped_cols
            if isinstance(f, types.BuiltinFunctionType):
                f = f.__name__
            if callable(f):
                lambda_name = _new_name("lambda")
                params = [(c.name, self.dtypes[c.name]) for c in c_]
                rtype = try_to_infer_type_of_operation(f, self.row_types)
                DaskContextManager.register_aggregation(
                    f, lambda_name, params, convert_type_to_pandas_dtype(rtype)
                )
                f_ = getattr(func, lambda_name)
            elif isinstance(f, str):
                if f == "first":
                    # first is registered as a postgresql function which
                    # behaves differently from pandas first.
                    # Use single_value instead.
                    f_ = getattr(func, "single_value")
                    rtype = self.row_types[src]
                elif f == "sum":
                    # sum is problematic since sqlalchemy has it return
                    # NullType so we force the return type to be the that
                    # of the src col.
                    f_ = getattr(func, f)
                    rtype = (
                        self.row_types[src]
                        if self.row_types is not None
                        else Unknown
                    )
                else:
                    f_ = getattr(func, f)
                    rtype = try_to_infer_type_of_operation(f, self.row_types)
            else:
                raise ValueError(
                    f"Aggregate function for {src} needs "
                    "to be callable or a string"
                )
            agg_cols.append(f_(*c_).label(str(dst)))
            agg_types[str(dst)] = rtype
        return agg_cols, agg_types

    def extended_projection(self, eval_expressions):
        if self.is_dee():
            return self._extended_projection_on_dee(eval_expressions)
        elif self._table is None:
            return type(self)(
                columns=list(eval_expressions.keys()), iterable=[]
            )

        proj_columns = []
        row_types = pd.Series(dtype="object")
        col_names = set()
        for dst_column, operation in eval_expressions.items():
            if callable(operation):
                lambda_name = _new_name("lambda")
                params = [(c, self.dtypes[c]) for c in self.columns]
                rtype = try_to_infer_type_of_operation(
                    operation, self.row_types
                )
                DaskContextManager.register_function(
                    operation,
                    lambda_name,
                    params,
                    convert_type_to_pandas_dtype(rtype),
                    True,
                )
                f_ = getattr(func, lambda_name)
                proj_columns.append(
                    f_(*self.sql_columns).label(str(dst_column))
                )
                row_types[str(dst_column)] = rtype
            elif isinstance(operation, RelationalAlgebraStringExpression):
                if str(operation) != str(dst_column):
                    proj_columns.append(
                        literal_column(operation).label(str(dst_column))
                    )
                    rtype = try_to_infer_type_of_operation(
                        operation, self.row_types
                    )
                    row_types[str(dst_column)] = rtype
                else:
                    proj_columns.append(self.sql_columns.get(str(operation)))
                    row_types[str(dst_column)] = self.row_types[
                        str(dst_column)
                    ]
                    col_names.add(str(operation))
            elif isinstance(operation, abc.RelationalAlgebraColumn):
                proj_columns.append(
                    self.sql_columns.get(str(operation)).label(str(dst_column))
                )
                row_types[str(dst_column)] = self.row_types[str(operation)]
                col_names.add(str(operation))
            else:
                proj_columns.append(literal(operation).label(str(dst_column)))
                rtype = try_to_infer_type_of_operation(
                    operation, self.row_types
                )
                row_types[str(dst_column)] = rtype

        query = select(proj_columns).select_from(self._table)
        if set(self.columns) != col_names:
            query = query.distinct()
        return self._create_view_from_query(
            query, row_types=row_types, is_empty=self._is_empty
        )

    def _extended_projection_on_dee(self, eval_expressions):
        """
        Extended projection when called on Dee to create set with
        constant values.
        """
        return type(self)(
            columns=eval_expressions.keys(),
            iterable=[eval_expressions.values()],
        )

    def to_unnamed(self):
        if self._table is not None:
            query = select(
                *[c.label(str(i)) for i, c in enumerate(self.sql_columns)]
            ).select_from(self._table)
            row_types = self.row_types.rename(
                {c: str(i) for i, c in enumerate(self.columns)}
            )
            return RelationalAlgebraFrozenSet()._create_view_from_query(
                query, row_types=row_types, is_empty=self._is_empty
            )
        return RelationalAlgebraFrozenSet(
            columns=[str(i) for i in range(self.arity)]
        )

    def projection_to_unnamed(self, *columns):
        unnamed_self = self.to_unnamed()
        named_columns = self.columns
        columns = tuple(named_columns.index(c) for c in columns)
        return unnamed_self.projection(*columns)


class RelationalAlgebraSet(
    RelationalAlgebraFrozenSet, abc.RelationalAlgebraSet
):
    def _update_self_with_ddf(
        self, ddf, _count=None, _is_empty=None, reset_row=False
    ):
        self._set_container(ddf, persist=True)
        self._count = _count
        self._is_empty = _is_empty
        if reset_row and hasattr(self, "_one_row"):
            delattr(self, "_one_row")

    def add(self, value):
        if self.container is None:
            value = [self._normalise_row(value)]
            self._create_insert_table(value, self._init_columns)
        else:
            value = self._normalise_element(value)
            ddf = self.container.append(pd.DataFrame([value], index=[0]))
            ddf = ddf.drop_duplicates()
            self._update_self_with_ddf(ddf, _is_empty=False)

    @staticmethod
    def _normalise_row(value):
        if isinstance(value, tuple):
            pass
        elif hasattr(value, "__iter__"):
            value = tuple(value)
        else:
            value = (value,)
        return value

    def discard(self, value):
        if self.container is not None:
            value = self._normalise_element(value)
            mask = (
                self.container[list(value.keys())] == list(value.values())
            ).all(axis=1)
            ddf = self.container[~mask]
            self._update_self_with_ddf(ddf)

    def __ior__(self, other):
        res = self.__or__(other)
        self._init_from(res)
        return self

    def __isub__(self, other):
        res = self.__sub__(other)
        self._init_from(res)
        return self



from typing import Callable
import numpy as np
from ..expressions import Constant

class NumpyFunctionsMixin():
    """
    This Mixin adds common mathematical functions in numpy to a
    NeurolangFrontend instance's DatalogProgram.

    The constant_X fields will be made available as functions to
    the frontend.
    """

    constant_exp = Constant[Callable[[float], float]](np.exp)
    constant_log = Constant[Callable[[float], float]](np.log)
    constant_log10 = Constant[Callable[[float], float]](np.log10)
    constant_cos = Constant[Callable[[float], float]](np.cos)
    constant_sin = Constant[Callable[[float], float]](np.sin)
    constant_tan = Constant[Callable[[float], float]](np.tan)


import logging
from concurrent.futures import Future, ThreadPoolExecutor
import types
from neurolang.commands import CommandsMixin
from neurolang.expressions import Command
from neurolang.frontend.query_resolution_expressions import Symbol
from neurolang.type_system import get_args, is_leq_informative
from threading import RLock, get_ident
from typing import AbstractSet, Callable, Dict, Union
from collections import OrderedDict

import nibabel
from neurolang.utils.relational_algebra_set import (
    NamedRelationalAlgebraFrozenSet,
    RelationalAlgebraFrozenSet,
)
from .engines import (
    NeurolangEngineConfiguration,
    NeurolangEngineSet,
)


LOG = logging.getLogger(__name__)


class LRUCacheDict(OrderedDict):
    """Dict with a limited length, ejecting LRUs as needed."""

    def __init__(self, cache_len: int = 15, *args, **kwargs):
        assert cache_len > 0
        self.cache_len = cache_len

        super().__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        super().move_to_end(key)

        while len(self) > self.cache_len:
            oldkey = next(iter(self))
            super().__delitem__(oldkey)

    def __getitem__(self, key):
        val = super().__getitem__(key)
        super().move_to_end(key)
        return val


class NeurolangQueryManager:
    """
    Class to manage execution of queries and keep track of results.

    This class is initialized with a dict of `NeurolangEngineConfiguration`
    that defines which Neurolang engine to create and how many.
    The NeurolangQueryManager creates a pool of thread workers (as many as
    there are engines) to execute the queries that are submited.

    It also keeps track of results in memory, in a results_cache which is a
    dict of uuid -> Future.
    """

    def __init__(
        self,
        options: Dict[NeurolangEngineConfiguration, int],
        cache_len: int = 15,
    ) -> None:
        """
        By default the `NeurolangQueryManager` creates
        `nb_engines = sum(options.values())` workers, but no engines. It will
        call the `_init_engines` method to create the engines asynchronously.

        Parameters
        ----------
        options : Dict[NeurolangEngineConfiguration, int]
            a dictionnary defining the types of engines and the number of each
            type to create.
        cache_len : int, optional
            the maximum size for the results cache, by default 15
        """
        self.engines = {}
        self._lock = RLock()
        self.results_cache = LRUCacheDict(cache_len=cache_len)
        self.configs = options

        nb_engines = sum(options.values())
        LOG.debug("Creating query manager with %d workers.", nb_engines)
        self.executor = ThreadPoolExecutor(max_workers=nb_engines)
        self._init_engines(options)

    def _init_engines(
        self, options: Dict[NeurolangEngineConfiguration, int]
    ) -> None:
        """
        Dispatch a series of tasks to create `nb_engines` engines using the
        `NeurolangEngineConfiguration.create` functions.

        Parameters
        ----------
        options : Dict[NeurolangEngineConfiguration, int]
            a dictionnary defining the types of engines and the number of each
            type to create.
        """

        def create_wrapper(config: NeurolangEngineConfiguration):
            engine = config.create()
            return (config.key, engine)

        # Dispatch the engine create tasks to the workers
        for config, nb in options.items():
            LOG.debug("Starting creation of %d %s engines...", nb, config.key)
            for _ in range(nb):
                future = self.executor.submit(create_wrapper, config)
                future.add_done_callback(self._engine_created)

    def _engine_created(self, future: Future) -> None:
        """
        Callback called when an engine creation task has finished.

        Parameters
        ----------
        future : Future
            the future holding the result of the engine creation task.
        """
        key, engine = future.result()
        with self._lock:
            engine_set = self.engines.get(key, None)
            if engine_set is None:
                engine_set = NeurolangEngineSet(engine)
                self.engines[key] = engine_set
            else:
                engine_set.add_engine(engine)
            LOG.debug(
                "Added a created engine of type %s, got %d of this type.",
                key, engine_set.counter
            )

    def _execute_neurolang_query(
        self, query: str, engine_type: str
    ) -> Dict[str, NamedRelationalAlgebraFrozenSet]:
        """
        Function executed on a ThreadPoolExecutor worker to execute a query
        against a neurolang engine.

        Parameters
        ----------
        query : str
            the query to execute
        engine_type : str
            the type of engine on which to execute the query

        Returns
        -------
        Dict[str, NamedRelationalAlgebraFrozenSet]
            the result of the query execution
        """
        LOG.debug("[Thread - %s] - Executing query...", get_ident())
        LOG.debug("[Thread - %s] - Query :\n%s", get_ident(), query)
        engine_set = self.engines[engine_type]
        with engine_set.engine() as engine:
            LOG.debug(
                "[Thread - %s] - Engine of type %s acquired.",
                get_ident(), engine_type
            )
            try:
                with engine.scope:
                    LOG.debug("[Thread - %s] - Solving query...", get_ident())
                    res = engine.execute_datalog_program(query)
                    if res is None:
                        res = engine.solve_all()
                    else:
                        res = {"ans": res}
                    LOG.debug("[Thread - %s] - Query solved.", get_ident())
                    for s in engine.program_ir.probabilistic_predicate_symbols:
                        try:
                            res[s]._is_probabilistic = True
                        except KeyError:
                            pass
                    if engine.current_program:
                        try:
                            last_parsed_symbol = engine.current_program[-1].expression.consequent.functor
                            res[last_parsed_symbol.name]._last_parsed_symbol = True
                        except Exception:
                            pass
                    return res
            except Exception as e:
                LOG.debug(
                    "[Thread - %s] - Query execution raised %s.", get_ident(), e
                )
                raise e
            finally:
                LOG.debug(
                    "[Thread - %s] - Engine of type %s released.", get_ident(), engine_type
                )

    def submit_query(self, uuid: str, query: str, engine_type: str) -> Future:
        """
        Submit a query to one of the available engines / workers.

        Parameters
        ----------
        uuid : str
            the uuid for the query.
        query : str
            the datalog query to execute.
        engine_type : str
            the type of engine to use for solving the query.

        Returns
        -------
        Future
            a future result for the query execution.
        """
        LOG.debug(
            "Submitting query with uuid %s to executor pool of %s engines.",
            uuid, engine_type
        )

        future_res = self.executor.submit(
            self._execute_neurolang_query, query, engine_type
        )
        self.results_cache[uuid] = future_res
        return future_res

    def get_result(self, uuid: str) -> Future:
        """
        Fetch the results for a query execution.

        Parameters
        ----------
        uuid : str
            the query id

        Returns
        -------
        Future
            the Future results of the execution task.
        """
        return self.results_cache[uuid]

    def cancel(self, uuid: str) -> bool:
        """
        Attempt to cancel the execution. If the call is currently being
        executed or finished running and cannot be cancelled then the method
        will return False, otherwise the call will be cancelled and the
        method will return True.

        Parameters
        ----------
        uuid : str
            the task execution id

        Returns
        -------
        bool
            True if cancelled
        """
        return self.results_cache[uuid].cancel()

    def get_atlas(self, engine_type: str) -> nibabel.Nifti1Image:
        """
        Get the atlas for a given engine type.

        Parameters
        ----------
        engine_type : str
            the engine type

        Returns
        -------
        nibabel.Nifti1Image
            the mni atlas
        """
        config = [c for c in self.configs.keys() if c.key == engine_type]
        return config[0].atlas

    def get_symbols(self, engine_type: str) -> Future:
        """
        Request the symbols table for a given engine. If the symbols table for
        this engine type is present in cache, then this is returned, otherwise
        a job to fetch the symbols table is dispatched to the threadpool
        executor.

        Parameters
        ----------
        engine_type : str
            the engine type.

        Returns
        -------
        Future
            the Future result for the symbols query
        """
        LOG.debug("Requesting symbols for %s engine...", engine_type)
        key = f"ENGINE_SYMBOLS_{engine_type}"
        if key in self.results_cache:
            LOG.debug("Returning cached symbols for %s engine.", engine_type)
            return self.results_cache[key]

        # Raise a KeyError if engines are not yet available
        self.engines[engine_type]

        # Submit task
        future_res = self.executor.submit(
            self._get_engine_symbols, engine_type
        )
        self.results_cache[key] = future_res
        return future_res

    def _get_engine_symbols(
        self, engine_type: str
    ) -> Dict[str, Union[RelationalAlgebraFrozenSet, Symbol]]:
        """
        Function executed on a ThreadPoolExecutor worker to get the available symbols
        on a neurolang engine.

        Parameters
        ----------
        engine_type : str
            the type of engine for which to get the symbols

        Returns
        -------
        Dict[str, Union[NamedRelationalAlgebraFrozenSet, Symbol]]
            the result of the query execution
        """
        LOG.debug(
            "[Thread - %s] - Fetching symbols for %s engine...",
            get_ident(), engine_type
        )
        engine_set = self.engines[engine_type]
        with engine_set.engine() as engine:
            try:
                LOG.debug(
                    "[Thread - %s] - Engine of type %s acquired.",
                    get_ident(), engine_type
                )
                LOG.debug(
                    "[Thread - %s] - Returning symbol_table.",
                    get_ident()
                )
                symbols = {}
                for name in engine.symbols:
                    if not name.startswith("_"):
                        symbol = engine.symbols[name]
                        if is_leq_informative(symbol.type, Callable):
                            symbols[name] = symbol
                        elif is_leq_informative(symbol.type, AbstractSet):
                            ras = symbol.value
                            ras.row_type = get_args(symbol.type)[0]
                            symbols[name] = ras
                symbols.update(_get_commands(engine))
                return symbols
            finally:
                LOG.debug(
                    "[Thread - %s] - Engine of type %s released.",
                    get_ident(), engine_type
                )


def _get_commands(engine) -> Dict[str, Dict]:
    """
    Add the command processing functions in CommandsMixin to the list
    of symbols available on an engine. This is useful to display the
    docstring of theses commands in the frontend.

    Parameters
    ----------
    engine : NeurolangPDL
        the engine with a program that extends CommandsMixin

    Returns
    -------
    Dict[str, Dict]
        a mapping of command name -> command info
    """
    commands = {}
    for name, command in CommandsMixin.__dict__.items():
        if (
            not name.startswith("_")
            and isinstance(command, types.FunctionType)
            and hasattr(engine.program_ir, name)
        ):
            commands[name] = {
                "type": str(Command),
                "doc": command.__doc__,
                "command": True,
            }
    return commands



"""
    Implementation of the weighted model counting approach through knowledge
    compilation.
"""
import logging
import operator as op
from typing import AbstractSet

import numpy as np
import pandas as pd
from pysdd import sdd

from ..config import config
from ..datalog.expression_processing import (
    extract_logic_predicates,
    flatten_query
)
from ..datalog.translate_to_named_ra import TranslateToNamedRA
from ..exceptions import NeuroLangException
from ..expression_walker import ExpressionWalker, PatternWalker, add_match
from ..expressions import (
    Constant,
    ExpressionBlock,
    FunctionApplication,
    Symbol,
    sure_is_not_pattern
)
from ..logic import FALSE, Implication
from ..relational_algebra import (
    ColumnInt,
    ColumnStr,
    ExtendedProjection,
    FunctionApplicationListMember,
    NameColumns,
    NumberColumns,
    Projection,
    PushInSelections,
    RelationalAlgebraStringExpression,
    RenameOptimizations,
    int2columnint_constant,
    str2columnstr_constant
)
from ..relational_algebra_provenance import (
    ProvenanceAlgebraSet,
    RelationalAlgebraProvenanceExpressionSemringSolver
)
from ..utils import log_performance
from ..utils.relational_algebra_set import NamedRelationalAlgebraFrozenSet
from .expression_processing import lift_optimization_for_choice_predicates
from .probabilistic_ra_utils import (
    DeterministicFactSet,
    ProbabilisticChoiceSet,
    ProbabilisticFactSet,
    generate_probabilistic_symbol_table_for_query
)
from .query_resolution import lift_solve_marg_query as _solve_marg_query

LOG = logging.getLogger(__name__)

ADD = Constant(op.add)
MUL = Constant(op.mul)
NEG = Constant(op.neg)


class SemiRingRAPToSDD(PatternWalker):
    def __init__(self, var_count, symbols_to_literals=None):
        self._init_manager(var_count)
        if symbols_to_literals is None:
            self.symbols_to_literals = dict()
        else:
            self.symbols_to_literals = symbols_to_literals
        self.literals_to_symbols = dict()

    def _init_manager(self, var_count):
        if isinstance(var_count, sdd.SddManager):
            self.manager = var_count.copy(list(var_count.vars))
        if isinstance(var_count, sdd.Vtree):
            self.manager = sdd.SddManager.from_vtree(var_count)
        else:
            self.manager = sdd.SddManager(var_count=var_count)

    @add_match(Symbol)
    def symbol(self, expression):
        if expression not in self.symbols_to_literals:
            var_num = len(self.symbols_to_literals) + 1
            self.symbols_to_literals[expression] = var_num
            self.literals_to_symbols[var_num] = expression

        lit = self.symbols_to_literals[expression]
        sdd_literal = self.manager.literal(lit)
        return sdd_literal

    @add_match(FunctionApplication(Constant(op.mul), ...))
    def mul(self, expression):
        args = list(expression.args)
        exp = self.manager.true()
        while len(args) > 0:
            arg = args.pop()
            if (
                isinstance(arg, FunctionApplication) and
                arg.functor == MUL
            ):
                args += arg.args
            else:
                arg = self.walk(arg)
                exp = exp & arg
        return exp

    @add_match(FunctionApplication(Constant(op.add), ...))
    def add(self, expression):
        args = self.walk(expression.args)
        args = list(expression.args)
        exp = self.manager.false()
        while len(args) > 0:
            arg = args.pop()
            if (
                isinstance(arg, FunctionApplication) and
                arg.functor == ADD
            ):
                args += arg.args
            else:
                arg = self.walk(arg)
                exp = exp | arg
        return exp

    @add_match(FunctionApplication(Constant(op.neg), ...))
    def neg(self, expression):
        arg = self.walk(expression.args[0])
        return ~arg

    @add_match(FunctionApplication(Constant(op.eq), ...))
    def eq(self, expression):
        args = self.walk(expression.args)
        return args[0].equiv(args[1])

    @add_match(FunctionApplication(Constant(op.or_), ...))
    def or_(self, expression):
        args = self.walk(expression.args)
        ret = args[0].condition(args[1])
        return ret

    @add_match(ExpressionBlock)
    def expression_block(self, expression):
        res = tuple(
            exp
            for exp in self.walk(expression.expressions)
        )
        for s in res:
            s.ref()
        return res


class EliminateSuperfluousProjectionMixin(PatternWalker):
    @add_match(
        Projection,
        lambda exp: (
            isinstance(
                exp.relation,
                (
                    DeterministicFactSet,
                    ProbabilisticFactSet,
                    ProbabilisticChoiceSet
                )
            )
            and all(att.type is ColumnInt for att in exp.attributes)
        )
    )
    def eliminate_superfluous_projection(self, expression):
        relation = self.walk(expression.relation)
        return relation


class DeterministicFactSetTranslation(PatternWalker):
    @add_match(
        DeterministicFactSet(Constant),
        lambda e: e.relation.value.is_empty()
    )
    def deterministic_fact_set_constant(self, deterministic_set):
        return ProvenanceAlgebraSet(
            deterministic_set.relation,
            int2columnint_constant(0)
        )


class WMCSemiRingSolver(
    EliminateSuperfluousProjectionMixin,
    DeterministicFactSetTranslation,
    RelationalAlgebraProvenanceExpressionSemringSolver,
):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.translated_probfact_sets = dict()
        self.tagged_sets = []

    def _semiring_mul(self, left, right):
        return FunctionApplication(Constant(self._internal_mul), (left, right))

    @staticmethod
    def _internal_mul(left, right):
        r = left * right
        return r

    @add_match(DeterministicFactSet(Symbol))
    def deterministic_fact_set(self, deterministic_set):
        relation_symbol = deterministic_set.relation
        if relation_symbol in self.translated_probfact_sets:
            return self.translated_probfact_sets[relation_symbol]

        with log_performance(
            LOG, f"Build deterministic fact set {relation_symbol}"
        ):
            relation = self.walk(relation_symbol)
            df = relation.value.as_pandas_dataframe()
            tags = [Symbol.fresh() for _ in range(len(df))]
            probs = [1.0 for _ in range(len(df))]
            tag_set = list(zip(tags, probs))
            prob_column = Symbol.fresh().name
            return self._build_tag_and_prov_sets(
                relation, relation_symbol, df, prob_column, tag_set, tags
            )

    @add_match(ProbabilisticFactSet(Symbol, Constant))
    def probabilistic_fact_set(self, prob_fact_set):
        relation_symbol = prob_fact_set.relation
        prob_column = prob_fact_set.probability_column.value
        if relation_symbol in self.translated_probfact_sets:
            return self.translated_probfact_sets[relation_symbol]

        with log_performance(
            LOG, f"Build probabilistic fact set {relation_symbol}"
        ):
            relation = self.walk(relation_symbol)
            df = relation.value.as_pandas_dataframe()
            tags = [Symbol.fresh() for _ in range(len(df))]
            probs = df[prob_column]
            tag_set = list(zip(tags, probs))
            return self._build_tag_and_prov_sets(
                relation, relation_symbol, df, prob_column, tag_set, tags
            )

    @add_match(ProbabilisticChoiceSet(Symbol, Constant))
    def probabilistic_choice_set(self, prob_choice_set):
        relation_symbol = prob_choice_set.relation
        prob_column = prob_choice_set.probability_column.value
        if relation_symbol in self.translated_probfact_sets:
            return self.translated_probfact_sets[relation_symbol]

        with log_performance(
            LOG, f"Build Probabilistic choice set {relation_symbol}"
        ):
            relation = self.walk(relation_symbol)
            df = relation.value.as_pandas_dataframe()
            previous = tuple()
            previous_probability = 0
            prov_column = []
            tag_set = []
            with sure_is_not_pattern():
                for probability in df[prob_column]:
                    adjusted_probability = (
                        probability / (1 - previous_probability)
                    )
                    previous_probability += probability
                    previous = self._generate_tag_choice_expression(
                        previous, adjusted_probability, prov_column, tag_set
                    )

            return self._build_tag_and_prov_sets(
                relation,
                relation_symbol,
                df,
                prob_column,
                tag_set,
                prov_column,
            )

    def _build_tag_and_prov_sets(
        self, relation, relation_symbol, df, prob_column, tag_set, prov_column
    ):
        tag_set = self._build_relation_constant(
            NamedRelationalAlgebraFrozenSet(("id", "prob"), tag_set)
        )
        self.tagged_sets.append(tag_set)
        tagged_df = df.copy()
        new_columns = tuple(f"col_{c}" for c in relation.value.columns)
        if prob_column not in df.columns:
            new_columns = new_columns + (f"col_{prob_column}",)
        tagged_df[prob_column] = prov_column
        tagged_ras_set = NamedRelationalAlgebraFrozenSet(
            new_columns, tagged_df
        )
        prov_set = ProvenanceAlgebraSet(
            Constant[AbstractSet](tagged_ras_set),
            str2columnstr_constant(f"col_{prob_column}")
        )
        self.translated_probfact_sets[relation_symbol] = prov_set
        return prov_set

    def _generate_tag_choice_expression(
        self, previous, adjusted_probability, prov_column, tag_set
    ):
        symbol = Symbol.fresh()
        neg_symbol = FunctionApplication(
            NEG, (symbol,), validate_arguments=False
        )
        if len(previous) == 0:
            tag_expression = symbol
            previous = (neg_symbol,)
        else:
            args = (symbol,) + previous
            tag_expression = FunctionApplication(
                MUL, args, validate_arguments=False
            )
        previous = (neg_symbol,) + previous
        prov_column.append(tag_expression)
        tag_set.append((symbol, adjusted_probability))
        return previous

    @add_match(ProbabilisticFactSet)
    def probabilistic_fact_set_invalid(self, prob_fact_set):
        raise NotImplementedError()


class SDDWMCSemiRingSolver(
    EliminateSuperfluousProjectionMixin,
    RenameOptimizations,
    DeterministicFactSetTranslation,
    RelationalAlgebraProvenanceExpressionSemringSolver,
):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.translated_probfact_sets = dict()
        self.tagged_sets = []
        self.var_count = 0
        symbols_seen = set()
        for v in self.symbol_table.values():
            if not isinstance(
                v,
                (
                    DeterministicFactSet,
                    ProbabilisticFactSet,
                    ProbabilisticChoiceSet
                )
            ):
                continue
            symbol = v.relation
            if symbol in symbols_seen:
                continue
            n = len(self.symbol_table[symbol].value)
            self.var_count += n
            symbols_seen.add(symbol)
        self.manager = sdd.SddManager(
            var_count=self.var_count,
            auto_gc_and_minimize=False
        )
        self.symbols_to_literals = dict()
        self.literals_to_symbos = dict()
        self.var_count = self.manager.var_count()
        self.positive_weights = []
        self._current_var = 1

    def wmc_weights(self):
        if len(self.positive_weights) != self.var_count:
            raise NeuroLangException("Not all SDD variables have been used")
        pos_weights = np.array(self.positive_weights)
        return np.r_[
            1 - pos_weights[::-1],
            pos_weights
        ]

    def get_new_bernoulli_variable(self, probability):
        literal = self.manager.literal(self._current_var)
        self._current_var += 1
        self.positive_weights.append(probability)
        return literal

    def _semiring_agg_sum(self, args):
        return FunctionApplication(
            Constant(self._internal_sum),
            args,
            validate_arguments=False,
            verify_type=False,
        )

    def _internal_sum(self, x):
        sum_ = self.manager.false()
        for el in x:
            el.ref()
            sum_ = sum_ | el
            el.deref()
        sum_.ref()
        return sum_

    def _semiring_mul(self, left, right):
        return FunctionApplication(
            Constant(self._internal_mul),
            (left, right)
        )

    @staticmethod
    def _internal_mul(left, right):
        left.ref()
        right.ref()
        r = left & right
        r.ref()
        left.deref()
        right.deref()

        return r

    @add_match(DeterministicFactSet(Symbol))
    def deterministic_fact_set(self, deterministic_set):
        relation_symbol = deterministic_set.relation
        if relation_symbol in self.translated_probfact_sets:
            return self.translated_probfact_sets[relation_symbol]

        relation = self.walk(relation_symbol)
        rap_column = Symbol.fresh().name

        new_columns = tuple(
            str2columnstr_constant(str(i))
            for i in relation.value.columns
        )

        relation = NameColumns(relation, new_columns)

        rap_column = str2columnstr_constant(Symbol.fresh().name)
        projection_list = tuple(
            FunctionApplicationListMember(
                Constant[RelationalAlgebraStringExpression](
                    RelationalAlgebraStringExpression(c.value),
                    verify_type=False
                ),
                c
            )
            for c in new_columns
        )
        deterministic_tag_function = FunctionApplication(
            Constant(lambda: self.get_new_bernoulli_variable(1.)),
            tuple()
        )

        new_relation = NumberColumns(
            ExtendedProjection(
                relation,
                projection_list + (
                    FunctionApplicationListMember(
                        deterministic_tag_function,
                        rap_column
                    ),
                )
            ),
            (rap_column,) + new_columns
        )

        tagged_relation = self.walk(new_relation)
        self.tagged_sets.append(tagged_relation)

        prov_set = ProvenanceAlgebraSet(
            tagged_relation, int2columnint_constant(0)
        )

        self.translated_probfact_sets[relation_symbol] = prov_set
        return prov_set

    @add_match(ProbabilisticFactSet(Symbol, Constant))
    def probabilistic_fact_set(self, prob_fact_set):
        relation_symbol = prob_fact_set.relation
        if relation_symbol in self.translated_probfact_sets:
            return self.translated_probfact_sets[relation_symbol]

        with log_performance(
            LOG, f"Build probabilistic fact set {relation_symbol}"
        ):
            relation = self.walk(relation_symbol)
            named_columns = tuple(
                str2columnstr_constant(f'col_{i}')
                for i in relation.value.columns
            )
            new_columns = tuple(
                c for i, c in enumerate(named_columns)
                if i != prob_fact_set.probability_column.value
            )

            relation = NameColumns(
                relation, named_columns
            )

            rap_column = str2columnstr_constant(Symbol.fresh().name)
            projection_list = [
                FunctionApplicationListMember(
                    Constant[RelationalAlgebraStringExpression](
                        RelationalAlgebraStringExpression(c.value),
                        verify_type=False
                    ),
                    c
                )
                for c in new_columns
            ]
            probfact_tag_function = FunctionApplication(
                Constant(self.get_new_bernoulli_variable),
                (named_columns[prob_fact_set.probability_column.value],)
            )

            tagged_relation = self.walk(
                NumberColumns(
                    ExtendedProjection(
                        relation,
                        projection_list + [
                            FunctionApplicationListMember(
                                probfact_tag_function,
                                rap_column
                            )
                        ]
                    ),
                    (rap_column,) + new_columns
                )
            )
            self.tagged_sets.append(tagged_relation)

            prov_set = ProvenanceAlgebraSet(
                tagged_relation, int2columnint_constant(0)
            )

            self.translated_probfact_sets[relation_symbol] = prov_set
        return prov_set

    @add_match(ProbabilisticChoiceSet(Symbol, Constant))
    def probabilistic_choice_set(self, prob_fact_set):
        relation_symbol = prob_fact_set.relation
        prob_column = prob_fact_set.probability_column.value
        if relation_symbol in self.translated_probfact_sets:
            return self.translated_probfact_sets[relation_symbol]

        with log_performance(
            LOG, f"Build Probabilistic choice set {relation_symbol}"
        ):
            relation = self.walk(relation_symbol)
            df = relation.value.as_pandas_dataframe()
            previous_probability = 0
            previous_expression = self.manager.true()
            tag_expressions = []
            for probability in df[prob_column]:
                adjusted_probability = probability / (1 - previous_probability)
                previous_probability += probability
                previous_expression = self.generate_sdd_expression(
                    adjusted_probability, tag_expressions, previous_expression
                )

            tagged_df = df.copy()
            tagged_df[prob_column] = tag_expressions
            new_columns = tuple(f'col_{c}' for c in relation.value.columns)
            tagged_ras_set = NamedRelationalAlgebraFrozenSet(
                new_columns, tagged_df
            )
        return ProvenanceAlgebraSet(
            NumberColumns(
                Constant[AbstractSet](tagged_ras_set, verify_type=False),
                tuple(str2columnstr_constant(c) for c in new_columns)
            ),
            int2columnint_constant(prob_column)
        )

    def generate_sdd_expression(
        self, adjusted_probability, tag_expressions, previous_expression
    ):
        bv = self.get_new_bernoulli_variable(adjusted_probability)
        previous_expression.ref()
        exp = bv & previous_expression
        exp.ref()
        previous_expression &= ~bv
        previous_expression.ref()
        tag_expressions.append(exp)
        return previous_expression

    @add_match(ProbabilisticFactSet)
    def probabilistic_fact_set_invalid(self, prob_fact_set):
        raise NotImplementedError()


def generate_weights(symbol_probs, literals_to_symbols, extras=0):
    ixs = pd.Series(literals_to_symbols)
    probs = symbol_probs.loc[ixs, 'prob'].values
    nprobs = symbol_probs.loc[ixs, 'nprob'].values
    probs = np.r_[nprobs[::-1], [0] * extras, [1] * extras, probs]
    return probs


class RAQueryOptimiser(
    PushInSelections,
    ExpressionWalker
):
    pass


def _build_empty_result_set(variables_to_project):
    cols = tuple(v.value for v in variables_to_project)
    prov_col = ColumnStr(Symbol.fresh().name)
    cols += (prov_col,)
    return ProvenanceAlgebraSet(
        Constant[AbstractSet](NamedRelationalAlgebraFrozenSet(
            iterable=[], columns=cols)
        ),
        str2columnstr_constant(prov_col),
    )


def solve_succ_query_boolean_diagram(
    query_predicate, cpl_program, run_relational_algebra_solver=False
):
    """
    Obtain the solution of a SUCC query on a CP-Logic program.

    The SUCC query must take the form

        SUCC[ P(x) ]
    """
    flat_query, ra_query = _prepare_and_translate_query(
        query_predicate, cpl_program
    )

    if flat_query == FALSE:
        return ra_query

    with log_performance(LOG, "Run RAP query"):
        symbol_table = generate_probabilistic_symbol_table_for_query(
            cpl_program, flat_query
        )

        solver = WMCSemiRingSolver(symbol_table)
        prob_set_result = solver.walk(ra_query)

    with log_performance(LOG, "SDD Model count"):
        res, provenance_column = sdd_compilation_and_wmc(
            prob_set_result, solver
        )

    return ProvenanceAlgebraSet(
        Constant[AbstractSet](res),
        str2columnstr_constant(provenance_column)
    )


def sdd_compilation_and_wmc(prob_set_result, solver):
    sdd_compiler, sdd_program, prob_set_program = \
        sdd_compilation(prob_set_result)

    if len(prob_set_program._symbols) > 0:
        res, provenance_column = perform_wmc(
            solver, sdd_compiler, sdd_program,
            prob_set_program, prob_set_result
        )
    else:
        provenance_column = ColumnStr('prob')
        res = NamedRelationalAlgebraFrozenSet(
            columns=(str(provenance_column),)
        )

    return res, provenance_column


def solve_succ_query_sdd_direct(
    query_predicate, cpl_program,
    per_row_model=True, run_relational_algebra_solver=True
):
    """
    Obtain the solution of a SUCC query on a CP-Logic program.

    The SUCC query must take the form

        SUCC[ P(x) ]
    """
    flat_query, ra_query = _prepare_and_translate_query(
        query_predicate, cpl_program
    )

    if flat_query == FALSE:
        return ra_query

    with log_performance(LOG, "Run RAP query"):
        symbol_table = generate_probabilistic_symbol_table_for_query(
            cpl_program, flat_query
        )
        solver = SDDWMCSemiRingSolver(symbol_table)
        prob_set_result = solver.walk(ra_query)

    df = prob_set_result.relation.value.as_pandas_dataframe()
    if per_row_model:
        probabilities = sdd_solver_per_individual_row(
            solver, df[prob_set_result.provenance_column.value]
        )
    else:
        probabilities = sdd_solver_global_model(
            solver, df[prob_set_result.provenance_column.value]
        )

    df[prob_set_result.provenance_column.value] = probabilities

    new_ras = type(prob_set_result.relation.value)(
        prob_set_result.relation.value.columns,
        df
    )

    return ProvenanceAlgebraSet(
        Constant[AbstractSet](new_ras),
        prob_set_result.provenance_column
    )


def _prepare_and_translate_query(query_predicate, cpl_program):
    with log_performance(LOG, 'Preparing query'):
        conjunctive_query, variables_to_project = prepare_initial_query(
            query_predicate
        )

        flat_query = flatten_query(conjunctive_query, cpl_program)
        if len(cpl_program.pchoice_pred_symbs) > 0:
            flat_query = lift_optimization_for_choice_predicates(
                flat_query, cpl_program
            )

        if flat_query == FALSE:
            ra_query = _build_empty_result_set(variables_to_project)
        else:
            with log_performance(LOG, "Translation and lifted optimisation"):
                ra_query = TranslateToNamedRA().walk(flat_query)
                ra_query = Projection(ra_query, variables_to_project)
                ra_query = RAQueryOptimiser().walk(ra_query)
    return flat_query, ra_query


def sdd_solver_global_model(solver, set_probabilities):
    with log_performance(LOG, "Build global SAT problem"):
        rows, exclusive_clause, initial_var_count = \
            build_global_sdd_model_rows(
                solver,
                set_probabilities
            )

        while len(rows) > 1:
            new_rows = []
            node_sizes = 0
            nodes_processed = 0
            solver.manager.minimize()
            for i in range(0, len(rows) - 1, 2):
                node_sizes += rows[i].size() + rows[i + 1].size()
                nodes_processed += 2
                new_rows.append(rows[i].conjoin(rows[i + 1]))
                rows[i].deref()
                rows[i + 1].deref()
                new_rows[-1].ref()
            new_rows += rows[i + 2:]
            rows = new_rows
        model = rows[0].conjoin(exclusive_clause)
    with log_performance(LOG, "Minimize manager"):
        model.ref()
        solver.manager.minimize()

    with log_performance(LOG, "Model count"):
        probabilities = model_count_and_per_row_probability(
            model, solver, initial_var_count, set_probabilities.shape[0]
        )
    return probabilities


def sdd_solver_per_individual_row(solver, set_probabilities):
    probabilities = np.empty(set_probabilities.shape[0])
    new_rows = []
    with log_performance(LOG, "Minimize manager"):
        new_rows = set_probabilities
        solver.manager.minimize()
    with log_performance(LOG, "Model Count"):
        weights = solver.wmc_weights()
        for i, row in enumerate(new_rows):
            tt = [row]
            nm = solver.manager.copy(tt)
            tt[0].ref()
            nm.minimize()
            wmc = row.wmc(log_mode=False)
            wmc.set_literal_weights_from_array(weights)
            probabilities[i] = wmc.propagate()
    return probabilities


def prepare_initial_query(query_predicate):
    if isinstance(query_predicate, Implication):
        conjunctive_query = query_predicate.antecedent
        variables_to_project = tuple(
            str2columnstr_constant(s.name)
            for s in query_predicate.consequent.args
            if isinstance(s, Symbol)
        )
    else:
        conjunctive_query = query_predicate
        variables_to_project = tuple(
            str2columnstr_constant(s.name)
            for s in query_predicate._symbols
            if s not in (
                p.functor for p in
                extract_logic_predicates(query_predicate)
            )
        )
    return conjunctive_query, variables_to_project


def model_count_and_per_row_probability(model, solver, initial_var_count, n):
    probabilities = np.empty(n)
    wmc = model.wmc(log_mode=False)
    pos_weights = np.r_[
        solver.positive_weights,
    ]
    extra_ones = np.ones(solver.manager.var_count() - initial_var_count)
    weights = np.r_[
        extra_ones,
        1 - pos_weights[::-1],
        pos_weights,
        extra_ones
    ]
    wmc.set_literal_weights_from_array(weights)
    for i in range(len(probabilities)):
        lit = solver.manager.literal(
            initial_var_count + i + 1
        )
        probabilities[i] = wmc.literal_derivative(lit)

    return probabilities


def build_global_sdd_model_rows(solver, literal_probabilities):
    rows = []
    neg = solver.manager.true()
    exclusive_clause = solver.manager.false()
    initial_var_count = solver.manager.var_count()
    for sdd_ in literal_probabilities:
        solver.manager.add_var_after_last()
        new_literal = solver.manager.literal(solver.manager.var_count())
        clause = (new_literal).equiv(sdd_)
        sdd_.deref()
        clause.ref()
        rows.append(clause)
        exclusive_clause |= new_literal  # & exclusive_clause
        old_neg = neg
        neg = neg & ~new_literal
        neg.ref()
        old_neg.deref()
    exclusive_clause.ref()
    return rows, exclusive_clause, initial_var_count


def sdd_compilation(prob_set_result):
    result_symbols = (
        prob_set_result
        .relation
        .value
        .projection(prob_set_result.provenance_column.value)
    )
    if result_symbols.is_empty():
        result_symbols = tuple()
    else:
        result_symbols = tuple(
            result_symbols.
            as_pandas_dataframe()
            .iloc[:, 0]
        )

    prob_set_program = ExpressionBlock(result_symbols)
    if len(prob_set_program._symbols) > 0:
        sdd_compiler = SemiRingRAPToSDD(len(prob_set_program._symbols))
        sdd_program = sdd_compiler.walk(prob_set_program)
    else:
        sdd_compiler = None
        sdd_program = None
    return sdd_compiler, sdd_program, prob_set_program


def perform_wmc(
    solver, sdd_compiler, sdd_program,
    prob_set_program, prob_set_result
):
    ids_with_probs = generate_probability_table(solver)

    weights = generate_weights(
        ids_with_probs,
        sdd_compiler.literals_to_symbols
    )

    probs = dict()
    for i, sdd_exp in enumerate(sdd_program):
        wmc = sdd_exp.wmc(log_mode=False)
        wmc.set_literal_weights_from_array(weights)
        prob = wmc.propagate()
        probs[prob_set_program.expressions[i]] = prob

    provenance_column = 'prob'
    while str2columnstr_constant(provenance_column) in prob_set_result.columns():
        provenance_column += '_'
    provenance_column = ColumnStr(provenance_column)

    extended_projections = dict(
        [
            (c.value, RelationalAlgebraStringExpression(c.value))
            for c in prob_set_result.columns()
            if c != prob_set_result.provenance_column
        ] +
        [(
            provenance_column,
            lambda x: probs[x[prob_set_result.provenance_column.value]]
        )]
    )
    res = prob_set_result.relation.value.extended_projection(
        extended_projections
    )
    return res, provenance_column


def generate_probability_table(solver):
    ids_with_probs = []
    for ts in solver.tagged_sets:
        ts = ts.value.as_pandas_dataframe()
        if 'nprob' not in ts:
            ts = ts.eval('nprob = 1 - prob')
        ids_with_probs.append(ts)
    ids_with_probs = pd.concat(ids_with_probs, axis=0).set_index('id')
    return ids_with_probs


if 'RAS' in config and config["RAS"].get("backend", "pandas") == "dask":
    solve_succ_query = solve_succ_query_boolean_diagram
else:
    solve_succ_query = solve_succ_query_sdd_direct


def solve_marg_query(rule, cpl):
    """
    Solve a MARG query on a CP-Logic program.

    Parameters
    ----------
    query : Implication
        Consequent must be of type `Condition`.
        MARG query of the form `ans(x) :- P(x)`.
    cpl_program : CPLogicProgram
        CP-Logic program on which the query should be solved.

    Returns
    -------
    ProvenanceAlgebraSet
        Provenance set labelled with probabilities for each tuple in the result
        set.

    """
    return _solve_marg_query(rule, cpl, solve_succ_query)
